import traceback
from typing import Any, Dict, Optional

from qcelemental.models import AtomicInput


class QCEngineException(Exception):
    """
    Base QCEngine exception, should never be called explicitly.
    """

    error_type = "base_error"
    header = "QCEngine Base Error"

    def __init__(self, message: str):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.raw_message = message
        self.traceback = traceback.format_exc()

    @property
    def error_message(self) -> str:
        return f"{self.header}: {self.raw_message}"


class UnknownError(QCEngineException):
    """
    Unknown QCEngine error, the type was not able to be specified.
    """

    error_type = "unknown_error"
    header = "QCEngine Unknown Error"


class InputError(QCEngineException):
    """
    Incorrect user parameters, not recoverable. Also may indicate a version issue.
    """

    error_type = "input_error"
    header = "QCEngine Input Error"


class ResourceError(QCEngineException):
    """
    Not enough resources for computation such as not enough memory, cores, or disk was not available.
    Recoverable on different compute hardware.
    """

    error_type = "resource_error"
    header = "QCEngine Resource Error"


class ConvergenceError(QCEngineException):
    """
    Failed iteration convergence error, likely recoverable with tweaked parameters.
    """

    error_type = "convergence_error"
    header = "QCEngine Convergence Error"


class RandomError(QCEngineException):
    """
    Likely recoverable errors such as segmentation faults or disk io problems.
    """

    error_type = "random_error"
    header = "QCEngine Random Error"


class KnownErrorException(QCEngineException):
    """Base class for detecting errors in the output of a ProgramHarness

    Subclasses for this class must define `error_name` and `description`, which provide
    a short name for the error used in the error correction code and a human-readable
    description of the error and how it is corrected.

    Subclasses for this exception must fulfill the `detect_error` class method,
     which detects if an error has occurred and raises an exception if one was detected.
     When raising the exception, the `detect_error` routine can optionally provide details
     about the error that are needed to correction.
    """

    error_type = "known_error"
    header = "QCEngine Known Error"

    error_name: str  # Unique name for this error. Used in autodetection logic
    description: str  # Human-readable description of the error.
    details: Optional[Dict[str, Any]]  # Any details for this error. Used for user feedback and autocorrection logic

    def __init__(self, details: Optional[dict] = None):
        super().__init__(self.description)
        self.details = details

    @classmethod
    def detect_error(cls, outputs: Dict[str, str]):
        """Detect whether this operation throws an error
        Args:
            outputs (dict): Output files needed to check for errors
        Raises:
            KnownErrorException: Error, if one was detected
        """
        raise NotImplementedError()

    def create_keyword_update(self, input_data: AtomicInput) -> Dict[str, Any]:
        """Create an keyword used to the update the dictionary given observed error

        Parameters
        ----------
            input_data
                Input specification used to perform the calculation that failed

        Returns
        -------
            Dictionary used to update the keywords field of ``input_data``
        """
        raise NotImplementedError()


class SimpleKnownErrorException(KnownErrorException):
    """Subclass for errors with simple detection logic.

    Most useful for error types that do not need any additional details to correct."""

    @classmethod
    def detect_error(cls, outputs: Dict[str, str]):
        if cls._detect(outputs):
            raise cls()

    @classmethod
    def _detect(cls, outputs: Dict[str, str]) -> bool:
        """Detect whether an error is present"""
        raise NotImplementedError()



import json
import os
import subprocess
import sys
from typing import List

from qcelemental.models import AtomicInput, OptimizationInput

from qcengine import cli, get_molecule, util
from qcengine.testing import using


def run_qcengine_cli(args: List[str], stdin: str = None) -> str:
    """
    Runs qcengine via its CLI

    This method was chosen over the more sophisticated util.execute in order to get pytest-cov to work.
    For the same reason, qcengine is invoked as "python -m qcengine.cli" rather than "qcengine".

    Parameters
    ----------
    args: List[str]
        List of CLI arguments.
    stdin: Optional[str]
        Standard input for the process.

    Returns
    -------
    str
        QCEngine CLI standard output.
    """
    if stdin is not None:
        stdin = stdin.encode("utf-8")

    return subprocess.check_output([sys.executable, "-m", "qcengine"] + args, input=stdin)


def test_no_args():
    """Test for qcengine with no arguments"""
    try:
        run_qcengine_cli([])
    except subprocess.CalledProcessError as e:
        assert e.returncode == 1


def test_info():
    """Test for qcengine info"""
    outputs = []
    for arg in cli.info_choices:
        output = run_qcengine_cli(["info", arg])
        if arg not in {"all", "config"}:  # output of config changes call-to-call depending e.g. on mem available
            outputs.append(output)

    default_output = run_qcengine_cli(["info"])
    for output in outputs:
        assert output in default_output


@using("psi4")
def test_run_psi4(tmp_path):
    """Tests qcengine run with psi4 and JSON input"""

    def check_result(stdout):
        output = json.loads(stdout)
        assert output["provenance"]["creator"].lower() == "psi4"
        assert output["success"] is True

    inp = AtomicInput(molecule=get_molecule("hydrogen"), driver="energy", model={"method": "hf", "basis": "6-31G"})

    args = ["run", "psi4", inp.json()]
    check_result(run_qcengine_cli(args))

    args = ["run", "psi4", os.path.join(tmp_path, "input.json")]
    with util.disk_files({"input.json": inp.json()}, {}, cwd=tmp_path):
        check_result(run_qcengine_cli(args))

    args = ["run", "psi4", "-"]
    check_result(run_qcengine_cli(args, stdin=inp.json()))


@using("geometric")
@using("psi4")
def test_run_procedure(tmp_path):
    """Tests qcengine run-procedure with geometric, psi4, and JSON input"""

    def check_result(stdout):
        output = json.loads(stdout)
        assert output["provenance"]["creator"].lower() == "geometric"
        assert output["success"] is True

    inp = {
        "keywords": {"coordsys": "tric", "maxiter": 100, "program": "psi4"},
        "input_specification": {"driver": "gradient", "model": {"method": "HF", "basis": "sto-3g"}, "keywords": {}},
        "initial_molecule": get_molecule("hydrogen"),
    }
    inp = OptimizationInput(**inp)

    args = ["run-procedure", "geometric", inp.json()]
    check_result(run_qcengine_cli(args))

    args = ["run-procedure", "geometric", os.path.join(tmp_path, "input.json")]
    with util.disk_files({"input.json": inp.json()}, {}, cwd=tmp_path):
        check_result(run_qcengine_cli(args))

    args = ["run-procedure", "geometric", inp.json()]
    check_result(run_qcengine_cli(args, stdin=inp.json()))



from typing import Union, Dict, Any

from qcelemental.models import OptimizationInput, AtomicInput, OptimizationResult, Provenance

from qcengine.config import TaskConfig
from qcengine.exceptions import UnknownError, InputError
from qcengine.procedures.nwchem_opt.harvester import harvest_as_atomic_result
from qcengine.programs.nwchem.runner import NWChemHarness
from qcengine.procedures.model import ProcedureHarness


class NWChemDriverProcedure(ProcedureHarness):
    """Structural relaxation using NWChem's optimizer"""

    _defaults = {"name": "NWChemDriver", "procedure": "optimization"}

    class Config(ProcedureHarness.Config):
        pass

    def found(self, raise_error: bool = False) -> bool:
        nwc_harness = NWChemHarness()
        return nwc_harness.found(raise_error)

    def get_version(self) -> str:
        nwc_harness = NWChemHarness()
        return nwc_harness.get_version()

    def build_input_model(self, data: Union[Dict[str, Any], "OptimizationInput"]) -> OptimizationInput:
        return self._build_model(data, OptimizationInput)

    def compute(self, input_data: OptimizationInput, config: TaskConfig) -> "BaseModel":
        nwc_harness = NWChemHarness()
        self.found(raise_error=True)

        # Unify the keywords from the OptimizationInput and QCInputSpecification
        #  Optimization input will override, but don't tell users this as it seems unnecessary
        keywords = input_data.keywords.copy()
        keywords.update(input_data.input_specification.keywords)
        if keywords.get("program", "nwchem").lower() != "nwchem":
            raise InputError("NWChemDriver procedure only works with NWChem")

        # Add a flag to the atomic input that tells the NWChemHarness we are calling it from driver
        #  This is needed for the NWCHarness to make some changes to the input file
        input_data.input_specification.extras["is_driver"] = True

        # Make an atomic input
        atomic_input = AtomicInput(
            molecule=input_data.initial_molecule,
            driver="energy",
            keywords=keywords,
            **input_data.input_specification.dict(exclude={"driver", "keywords"}),
        )

        # Build the inputs for the job
        job_inputs = nwc_harness.build_input(atomic_input, config)

        # Replace the last line with a "task {} optimize"
        input_file: str = job_inputs["infiles"]["nwchem.nw"].strip()
        beginning, last_line = input_file.rsplit("\n", 1)
        assert last_line.startswith("task")
        last_line = f"task {last_line.split(' ')[1]} optimize"
        job_inputs["infiles"]["nwchem.nw"] = f"{beginning}\n{last_line}"

        # Run it!
        success, dexe = nwc_harness.execute(job_inputs)

        # Check for common errors
        if "There is an error in the input file" in dexe["stdout"]:
            raise InputError(dexe["stdout"])
        if "not compiled" in dexe["stdout"]:
            # recoverable with a different compilation with optional modules
            raise InputError(dexe["stdout"])

        # Parse it
        if success:
            dexe["outfiles"]["stdout"] = dexe["stdout"]
            dexe["outfiles"]["stderr"] = dexe["stderr"]
            return self.parse_output(dexe["outfiles"], input_data)
        else:
            raise UnknownError(dexe["stdout"])

    def parse_output(self, outfiles: Dict[str, str], input_model: OptimizationInput) -> OptimizationResult:

        # Get the stdout from the calculation (required)
        stdout = outfiles.pop("stdout")
        stderr = outfiles.pop("stderr")

        # Parse out the atomic results from the file
        atomic_results = harvest_as_atomic_result(input_model, stdout)

        # Isolate the converged result
        final_step = atomic_results[-1]

        return OptimizationResult(
            initial_molecule=input_model.initial_molecule,
            input_specification=input_model.input_specification,
            final_molecule=final_step.molecule,
            trajectory=atomic_results,
            energies=[float(r.extras["qcvars"]["CURRENT ENERGY"]) for r in atomic_results],
            stdout=stdout,
            stderr=stderr,
            success=True,
            provenance=Provenance(creator="NWChemRelax", version=self.get_version(), routine="nwchem_opt"),
        )



import numpy as np
import pytest
import qcelemental
from qcelemental.testing import compare_values

import qcengine as qcng
from qcengine.programs.turbomole.harvester import parse_hessian
from qcengine.testing import using


@pytest.fixture
def h2o():
    mol = qcelemental.models.Molecule.from_data(
        """
            O 0.000000000000     0.000000000000    -0.068516245955
            H 0.000000000000    -0.790689888800     0.543701278274
            H 0.000000000000     0.790689888800     0.543701278274
    """
    )
    return mol


@pytest.fixture
def h2o_ricc2_def2svp():
    """NumForce calls only make sense for stationary points. So this
    geometry was optimized at the ricc2/def2-svp level of theory and
    can be used to run NumForce with ricc2."""

    mol = qcelemental.models.Molecule.from_data(
        """
        O     0.0000000    0.0000000   -0.0835835
        H     0.7501772    0.0000000    0.5210589
        H    -0.7501772    0.0000000    0.5210589
    """
    )
    return mol


@pytest.mark.parametrize(
    "method, keywords, ref_energy",
    [
        pytest.param("hf", {}, -75.95536954370, marks=using("turbomole")),
        pytest.param("pbe0", {"grid": "m5"}, -76.27371135900, marks=using("turbomole")),
        pytest.param("ricc2", {}, -76.1603807755, marks=using("turbomole")),
        pytest.param("rimp2", {}, -76.1593614075, marks=using("turbomole")),
    ],
)
def test_turbomole_energy(method, keywords, ref_energy, h2o):
    resi = {"molecule": h2o, "driver": "energy", "model": {"method": method, "basis": "def2-SVP"}, "keywords": keywords}

    res = qcng.compute(resi, "turbomole", raise_error=True, return_dict=True)

    assert res["driver"] == "energy"
    assert res["success"] is True

    assert compare_values(ref_energy, res["return_result"])


@pytest.mark.parametrize(
    "method, keywords, ref_norm",
    [
        pytest.param("hf", {}, 0.099340, marks=using("turbomole")),
        pytest.param("pbe0", {"grid": "m5"}, 0.0606266, marks=using("turbomole")),
        pytest.param("ricc2", {}, 0.059378, marks=using("turbomole")),
        pytest.param("rimp2", {}, 0.061576, marks=using("turbomole")),
    ],
)
def test_turbomole_gradient(method, keywords, ref_norm, h2o):
    resi = {
        "molecule": h2o,
        "driver": "gradient",
        "model": {"method": method, "basis": "def2-SVP"},
        "keywords": keywords,
    }

    res = qcng.compute(resi, "turbomole", raise_error=True)

    assert res.driver == "gradient"
    assert res.success is True
    assert res.properties.return_energy

    grad = res.return_result
    grad_norm = np.linalg.norm(grad)
    assert compare_values(ref_norm, grad_norm)


@using("turbomole")
def test_turbomole_ri_dsp(h2o):
    resi = {
        "molecule": h2o,
        "driver": "energy",
        "model": {"method": "b-p", "basis": "def2-SVP"},
        "keywords": {"ri": True, "d3bj": True},
    }

    res = qcng.compute(resi, "turbomole", raise_error=True)

    assert res.driver == "energy"
    assert res.success is True

    energy = res.return_result
    ref_energy = -76.36275642866
    assert compare_values(ref_energy, energy)


def assert_hessian(H, ref_eigvals, ref_size):
    w, v = np.linalg.eigh(H)
    last_eigvals = w[-3:]
    # Hessian must be symmetric
    np.testing.assert_allclose(H, H.T)
    # Check eigenvalues
    np.testing.assert_allclose(last_eigvals, ref_eigvals)
    # Hessian must be of shape (3N x 3N)
    assert H.shape == (ref_size, ref_size)


@using("turbomole")
@pytest.mark.parametrize(
    "method, keywords, ref_eigvals",
    [
        ("hf", {}, (2.00771683e-01, 7.77977644e-01, 9.91091318e-01)),
        ("pbe0", {"grid": "m5"}, (1.72092719e-01, 7.38603449e-01, 9.73783598e-01)),
        ("b-p", {"grid": "m5", "ri": True}, (1.59729409e-01, 7.21364827e-01, 9.63399519e-01)),
    ],
)
def test_turbomole_hessian(method, keywords, ref_eigvals, h2o):
    resi = {
        "molecule": h2o,
        "driver": "hessian",
        "model": {
            "method": method,
            "basis": "def2-SVP",
        },
        "keywords": keywords,
    }

    res = qcng.compute(resi, "turbomole", raise_error=True)
    H = res.return_result
    size = h2o.geometry.size

    assert res.driver == "hessian"
    assert res.success is True
    assert res.properties.return_energy
    assert_hessian(H, ref_eigvals, size)


@using("turbomole")
@pytest.mark.parametrize(
    "method, keywords, ref_eigvals",
    [
        ("ricc2", {}, (1.65405531e-01, 9.63690706e-01, 1.24676634e00)),
    ],
)
def test_turbomole_num_hessian(method, keywords, ref_eigvals, h2o_ricc2_def2svp):
    resi = {
        "molecule": h2o_ricc2_def2svp,
        "driver": "hessian",
        "model": {
            "method": method,
            "basis": "def2-SVP",
        },
        "keywords": keywords,
    }

    res = qcng.compute(resi, "turbomole", raise_error=True)
    H = res.return_result

    size = h2o_ricc2_def2svp.geometry.size

    assert res.driver == "hessian"
    assert res.success is True
    assert res.properties.return_energy
    assert_hessian(H, ref_eigvals, size)


@pytest.fixture
def h2o_nprhessian():
    return """$nprhessian
     1  1   0.6142699252  -0.0000000000   0.0000000000  -0.3071349626  -0.2479448514
     1  2  -0.0000000000  -0.3071349626   0.2479448514  -0.0000000000
     2  1  -0.0000000000   0.4365036678   0.0000000000  -0.1885017686  -0.2182518339
     2  2  -0.0000000000   0.1885017686  -0.2182518339   0.0000000000
     3  1   0.0000000000   0.0000000000  -0.0000524175  -0.0000000000  -0.0000000000
     3  2   0.0000262088  -0.0000000000   0.0000000000   0.0000262088
     4  1  -0.3071349626  -0.1885017686  -0.0000000000   0.3389423895   0.2182233100
     4  2   0.0000000000  -0.0318074269  -0.0297215414  -0.0000000000
     5  1  -0.2479448514  -0.2182518339  -0.0000000000   0.2182233100   0.2092172237
     5  2   0.0000000000   0.0297215414   0.0090346102   0.0000000000
     6  1  -0.0000000000  -0.0000000000   0.0000262088   0.0000000000   0.0000000000
     6  2  -0.0000125560  -0.0000000000   0.0000000000  -0.0000136528
     7  1  -0.3071349626   0.1885017686  -0.0000000000  -0.0318074269   0.0297215414
     7  2  -0.0000000000   0.3389423895  -0.2182233100   0.0000000000
     8  1   0.2479448514  -0.2182518339   0.0000000000  -0.0297215414   0.0090346102
     8  2   0.0000000000  -0.2182233100   0.2092172237  -0.0000000000
     9  1  -0.0000000000   0.0000000000   0.0000262088  -0.0000000000   0.0000000000
     9  2  -0.0000136528   0.0000000000  -0.0000000000  -0.0000125560
    $end"""


def test_turbomole_parse_hessian(h2o_nprhessian):
    """Test parsing of unproject Turbomole Hessian for water."""
    hessian = parse_hessian(h2o_nprhessian)
    assert hessian.shape == (9, 9)
    eigvals, _ = np.linalg.eigh(hessian)
    assert eigvals[-1] == pytest.approx(1.12157030e00)



"""
Harness for the extended tight binding (xtb) package.
This implementation interfaces with the xtb C-API via the xtb-python project,
which is providing Python bindings and native extensions for common computational
chemistry frameworks, like QCEngine.

Therefore, this harness only has to provide a thin wrapper to integrate xtb.
For more information on xtb-python and the actual QCSchema integration,
visit `its documentation <https://xtb-python.readthedocs.io>`_.
"""

from typing import Dict

from qcelemental.models import AtomicInput, AtomicResult
from qcelemental.util import safe_version, which_import

from ..config import TaskConfig
from .model import ProgramHarness


class XTBHarness(ProgramHarness):
    """Calculation harness for the extended tight binding (xtb) package."""

    _defaults = {
        "name": "xtb",
        "scratch": False,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """Check for the availability of the Python API of xtb"""

        return which_import(
            "xtb",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install xtb-python -c conda-forge`.",
        )

    def get_version(self) -> str:
        """Return the currently used version of xtb-python"""
        self.found(raise_error=True)

        which_prog = which_import("xtb")
        if which_prog not in self.version_cache:
            import xtb

            self.version_cache[which_prog] = safe_version(xtb.__version__)

        return self.version_cache[which_prog]

    def compute(self, input_data: AtomicInput, config: TaskConfig) -> AtomicResult:
        """
        Actual interface to the xtb package. The compute function is just a thin
        wrapper around the native QCSchema interface of xtb-python.
        """

        self.found(raise_error=True)

        import xtb
        from xtb.qcschema.harness import run_qcschema

        # Run the Harness
        output = run_qcschema(input_data)

        # Make sure all keys from the initial input spec are sent along
        output.extras.update(input_data.extras)
        return output



from math import sqrt
import re
from decimal import Decimal

import numpy as np

from ..util import PreservingDict


def parse_decimal(regex, text, method="search"):
    with_float = re.compile(regex + "([\d\-\.]+)")
    matches = getattr(with_float, method)(text)

    if method == "search":
        matches = [matches.groups()]
    return [(method, Decimal(energy)) for method, energy in matches]


def parse_reference_energy(stdout: str):
    """Parse stdout and return the energy of the reference wavefunction."""
    energy_dict = PreservingDict()

    # Total energy from dscf or ridft
    total_energy_re = re.compile("total energy\s+=\s+([\d\-\.]+)")
    mobj = total_energy_re.search(stdout)
    total_energy = Decimal(mobj[1])

    # Check for DFT, default to HF
    energy_key = "HF TOTAL ENERGY"
    dft_mobj = re.search("density functional", stdout)
    if dft_mobj:
        energy_key = "DFT TOTAL ENERGY"
    energy_dict[energy_key] = total_energy

    # Take into account energies from ricc2 runs. They will be different
    # from the HF energy.
    current_energy = total_energy
    energy_dict["CURRENT ENERGY"] = current_energy

    return energy_dict


def parse_ricc2(stdout: str):
    ricc2_dict = PreservingDict()

    # As CC2 starts from a MP2 guess that is also reported there may be
    # multiple matches for the following regex. Thats why we capture all
    # matches with 'findall'.
    matches = parse_decimal("Final (.+?) energy\s+:\s+", stdout, "findall")
    if len(matches) == 0:
        matches = parse_decimal("E(MP2)\s+:\s+", stdout, "search")

    ricc2_dict["CURRENT ENERGY"] = matches[-1][1]

    return ricc2_dict


def parse_gradient(gradient):
    grad_re = re.compile(
        "\$grad.+"
        "cycle =\s+(?P<cycle>\d+)\s+"
        "(?P<energy_type>.+?) energy =\s+(?P<energy>[\d\.\-]+)\s+"
        "\|dE/dxyz\| =\s+(?P<grad_norm>[\d\.]+)"
        "(?P<coords_gradients>.+)\$end",
        re.DOTALL,
    )
    mobj = grad_re.match(gradient)

    # Commented out the variables that aren't returned so LGTM doesn't
    # complain.
    # energy_type = mobj.group("energy_type")
    # grad_norm = Decimal(mobj.group("grad_norm"))
    # energy = Decimal(mobj.group("energy"))
    coords_grad = mobj.group("coords_gradients")
    # cycle = int(mobj.group("cycle"))

    *_, grad = re.split("[a-z]{1,3}", coords_grad.strip())
    grad = np.array(grad.strip().replace("D", "E").split(), dtype=float)

    return grad


def parse_hessian(hessian):
    first_hessian = hessian.strip().split("$")[1]
    split = first_hessian.split()
    hess_type = split.pop(0)
    assert hess_type in ("nprhessian", "hessian")

    def is_float(str_):
        return "." in str_

    hess_items = [item for item in split if is_float(item)]
    coord_num = int(sqrt(len(hess_items)))
    assert coord_num**2 == len(hess_items)
    hessian = np.array(hess_items, dtype=float).reshape(-1, coord_num)

    return hessian


def harvest(input_model, stdout, **outfiles):
    qcvars = PreservingDict()

    ref_energy_dict = parse_reference_energy(stdout)
    qcvars.update(ref_energy_dict)

    if "R I C C 2" in stdout:
        ricc2_dict = parse_ricc2(stdout)
        qcvars.update(ricc2_dict)

    gradient = None
    if "gradient" in outfiles:
        gradient = parse_gradient(outfiles["gradient"])
        qcvars["N ATOMS"] = gradient.size // 3

    # Prefer unprojected 'nprhessian' over projected 'hessian'.
    hessian_text = outfiles.get("nprhessian", outfiles.get("hessian", None))

    if hessian_text is not None:
        hessian = parse_hessian(hessian_text)
        qcvars["N ATOMS"] = hessian.shape[0] // 3
    else:
        hessian = None

    return qcvars, gradient, hessian



import collections
from typing import Any, Dict, Tuple


def format_keyword(keyword: str, val: Any, lop_off: bool = True, preserve_case: bool = False) -> Tuple[str, str]:
    """Function to reformat value `val` for `keyword` from python into nwchem-speak."""

    if preserve_case:
        key = keyword
    else:
        key = keyword.lower()

    # Transform string booleans into " "
    if val is True:
        return key, ""
    elif val is False:
        return "", ""

    # complete hack
    if keyword.upper() == "MEMORY":
        return key, f"{val} double"

    elif isinstance(val, list):
        text = " ".join([str(v) for v in val])
    elif isinstance(val, dict):
        text = []
        for k, v in val.items():
            merge = [k]
            merge.extend(str(v) if isinstance(v, (int, float)) else list(map(str, v)))
            text.append(" ".join(merge))
        text = " ".join(text)
    else:
        text = str(val)

    if lop_off:
        return key[7:], text
    else:
        return key, text


def format_keywords(keywords: Dict[str, Any]) -> str:
    """From NWCHEM-directed, non-default `keywords` dictionary, write a NWCHEM deck."""

    def rec_dd():
        return collections.defaultdict(rec_dd)

    grouped_options = rec_dd()

    for group_key, val in keywords.items():
        nesting = group_key.split("__")
        if len(nesting) == 1:
            key = nesting[0]
            grouped_options["aaaglobal"][key] = val
        elif len(nesting) == 2:
            g1, key = nesting
            grouped_options[g1][key] = val
        elif len(nesting) == 3:
            g1, g2, key = nesting
            grouped_options[g1][g2][key] = val
        else:
            raise ValueError("Nesting N!")

    grouped_lines = {}
    for group, opts in sorted(grouped_options.items()):
        lines = []
        group_level_lines = []
        for key, val in grouped_options[group].items():
            if isinstance(val, dict):
                g2_level_lines = []
                g2_level_lines.append(key.lower())
                for k2, v2 in val.items():
                    line2 = " ".join(format_keyword(k2, v2, lop_off=False))
                    g2_level_lines.append(line2)
                g2_level_lines = " ".join(g2_level_lines)
                lines.append(g2_level_lines)
            else:
                preserve_case = False
                try:
                    preserve_case = val.startswith("library ")
                except AttributeError:
                    pass
                line = " ".join(format_keyword(key, val, lop_off=False, preserve_case=preserve_case))
                if group.lower() == "basis" and any(
                    [word in line for word in ["spherical", "cartesian", "print", "noprint", "rel"]]
                ):
                    group_level_lines.append(line)
                else:
                    lines.append(line)
        if group == "aaaglobal":
            grouped_lines[group] = "\n".join(lines) + "\n"
        elif group.lower() == "set":
            grouped_lines[group] = "\n".join(f"set {l}" for l in lines) + "\n"
        else:
            grouped_lines[group] = (
                f"{group.lower()} " + " ".join(group_level_lines) + "\n  " + "\n  ".join(lines) + "\nend\n"
            )

    return "\n".join(grouped_lines.values()) + "\n"



"""
Calls the Q-Chem executable.
"""

import os
import re
import tempfile
import warnings
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from qcelemental import constants
from qcelemental.models import AtomicInput, AtomicResult, Molecule, Provenance
from qcelemental.molparse import regex
from qcelemental.util import parse_version, safe_version, which

from qcengine.config import TaskConfig, get_config

from ..exceptions import InputError, UnknownError
from ..util import disk_files, execute, temporary_directory
from .model import ProgramHarness

NUMBER = r"(?x:" + regex.NUMBER + ")"


class QChemHarness(ProgramHarness):
    _defaults: Dict[str, Any] = {
        "name": "QChem",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    def found(self, raise_error: bool = False) -> bool:
        return which(
            "qchem",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install by visiting the Q-Chem website. Check it's in your PATH with `which qchem'.",
        )

    def _get_qc_path(self, config: Optional["TaskConfig"] = None):
        paths = os.environ.copy()
        paths["QCSCRATCH"] = tempfile.gettempdir()
        if config and config.scratch_directory:
            paths["QCSCRATCH"] = config.scratch_directory

        # Nothing else to pass in
        if "QC" in os.environ:
            return paths

        # Assume QC path is
        qchem_path = which("qchem")
        if qchem_path:
            paths["QC"] = os.path.dirname(os.path.dirname(qchem_path))

        return paths

    def get_version(self) -> str:
        # excuse missing Q-Chem for QCEngineRecords tests
        found = self.found()
        if not found:
            return safe_version("0.0.0")

        self.found(raise_error=True)

        # Get the node configuration
        config = get_config()

        which_prog = which("qchem")
        if which_prog not in self.version_cache:
            success, exc = execute(
                [which_prog, "v.in"],
                {"v.in": "$rem\n"},
                scratch_directory=config.scratch_directory,
                environment=self._get_qc_path(),
                timeout=15,
            )

            mobj = re.search(r"Q-Chem\s+([\d.]+)\s+for", exc["stdout"])
            if not mobj:
                mobj = re.search(r"Q-Chem version:\s+([\d.]+)\s+", exc["stdout"])

            if mobj:
                self.version_cache[which_prog] = safe_version(mobj.group(1))

            # if "QC not defined" in exc["stdout"]:
            else:
                return safe_version("0.0.0")

        return self.version_cache[which_prog]

    def compute(self, input_model: "AtomicInput", config: TaskConfig) -> "AtomicResult":
        """
        Run qchem
        """
        # Check if qchem executable is found
        self.found(raise_error=True)

        # Check qchem version
        qceng_ver = "5.1"
        if parse_version(self.get_version()) < parse_version(qceng_ver):
            raise TypeError(f"Q-Chem version <{qceng_ver} not supported (found version {self.get_version()})")

        # Setup the job
        job_inputs = self.build_input(input_model, config)

        # Run qchem
        exe_success, proc = self.execute(job_inputs)

        # Determine whether the calculation succeeded
        if exe_success:
            # If execution succeeded, collect results
            result = self.parse_output(proc["outfiles"], input_model)
            return result
        else:
            outfile = proc["outfiles"]["dispatch.out"]
            if "fatal error occurred in module qparser" in outfile:
                raise InputError(proc["outfiles"]["dispatch.out"])
            else:
                # Return UnknownError for error propagation
                raise UnknownError(proc["outfiles"]["dispatch.out"])

    def execute(
        self,
        inputs: Dict[str, Any],
        *,
        extra_infiles: Optional[Dict[str, str]] = None,
        extra_outfiles: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        scratch_messy: bool = False,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:
        """
        For option documentation go look at qcengine/util.execute
        """

        # Collect all input files and update with extra_infiles
        infiles = inputs["infiles"]
        if extra_infiles is not None:
            infiles.update(extra_infiles)

        binary_files = [os.path.join("savepath", x) for x in ["99.0", "131.0", "132.0"]]

        # Collect all output files and extend with with extra_outfiles
        outfiles = ["dispatch.out"]
        if extra_outfiles is not None:
            outfiles.extend(extra_outfiles)

        # Replace commands with extra_commands if present
        commands = inputs["commands"] + ["savepath"]
        if extra_commands is not None:
            commands = extra_commands

        envs = self._get_qc_path()

        with temporary_directory(parent=inputs["scratch_directory"], suffix="_qchem_scratch") as tmpdir:
            envs["QCSCRATCH"] = tmpdir
            bdict = {x: None for x in binary_files}

            with disk_files({}, bdict, cwd=tmpdir, as_binary=binary_files):
                exe_success, proc = execute(
                    commands,
                    infiles=infiles,
                    outfiles=outfiles,
                    scratch_name=scratch_name,
                    scratch_directory=tmpdir,
                    scratch_messy=scratch_messy,
                    timeout=timeout,
                    environment=envs,
                )

            proc["outfiles"].update({os.path.split(k)[-1]: v for k, v in bdict.items()})

        if (proc["outfiles"]["dispatch.out"] is None) or (
            "Thank you very much for using Q-Chem" not in proc["outfiles"]["dispatch.out"]
        ):
            exe_success = False

        # QChem does not create an output file and only prints to stdout
        return exe_success, proc

    def build_input(
        self, input_model: "AtomicInput", config: TaskConfig, template: Optional[str] = None
    ) -> Dict[str, Any]:

        # Check some bounds on what cannot be parsed
        if "ccsd" in input_model.model.method.lower() or "ccd" in input_model.model.method.lower():
            raise InputError("Cannot handle CC* methods currently.")

        # Build keywords
        keywords = {k.upper(): v for k, v in input_model.keywords.items()}
        keywords["INPUT_BOHR"] = "TRUE"
        keywords["MEM_TOTAL"] = str(int(config.memory * 1024))  # In MB

        if input_model.driver == "energy":
            keywords["JOBTYPE"] = "sp"
        elif input_model.driver == "gradient":
            keywords["JOBTYPE"] = "force"
        elif input_model.driver == "hessian":
            keywords["JOBTYPE"] = "freq"
        else:
            raise InputError(f"Driver {input_model.driver} not implemented for Q-Chem.")

        if input_model.molecule.fix_com or input_model.molecule.fix_orientation:
            keywords["SYM_IGNORE"] = "TRUE"

        keywords["METHOD"] = input_model.model.method
        if input_model.model.basis:
            keywords["BASIS"] = input_model.model.basis

        # Begin the input file
        input_file = []
        input_file.append(
            f"""$comment
Automatically generated Q-Chem input file by QCEngine
$end
            """
        )

        # Add Molecule, TODO: Add to QCElemental
        mol = input_model.molecule
        input_file.append("$molecule")
        input_file.append(f"""{int(mol.molecular_charge)} {mol.molecular_multiplicity}""")

        for real, sym, geom in zip(mol.real, mol.symbols, mol.geometry):
            if real is False:
                raise InputError("Cannot handle ghost atoms yet.")
            input_file.append(f"{sym} {geom[0]:14.8f} {geom[1]:14.8f} {geom[2]:14.8f}")

        input_file.append("$end\n")

        # Write out the keywords
        input_file.append("$rem")
        for k, v in keywords.items():
            input_file.append(f"{k:20s}  {v}")
        input_file.append("$end\n")

        ret = {
            "infiles": {"dispatch.in": "\n".join(input_file)},
            "commands": [which("qchem"), "-nt", str(config.ncores), "dispatch.in", "dispatch.out"],
            "scratch_directory": config.scratch_directory,
        }

        return ret

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> AtomicResult:

        output_data = {}

        bdata = {}
        outtext = ""
        for k, v in outfiles.items():
            if k == "dispatch.out":
                outtext = v
                continue
            if v is None:
                continue
            bdata[k] = np.frombuffer(v)

        if input_model.driver == "energy":
            output_data["return_result"] = bdata["99.0"][-1]
        elif input_model.driver == "gradient":
            output_data["return_result"] = bdata["131.0"]
        elif input_model.driver == "hessian":
            output_data["return_result"] = bdata["132.0"]
        else:
            raise ValueError(f"Could not parse driver of type {input_model.driver}.")

        properties = {
            "nuclear_repulsion_energy": bdata["99.0"][0],
            "scf_total_energy": bdata["99.0"][1],
            "return_energy": bdata["99.0"][-1],
        }

        qcvars = {}
        _mp2_methods = {"mp2", "rimp2"}
        if input_model.model.method.lower() in _mp2_methods:
            emp2 = bdata["99.0"][-1]
            properties["mp2_total_energy"] = emp2
            qcvars["MP2 TOTAL ENERGY"] = emp2
            qcvars["CURRENT ENERGY"] = emp2
            ehf = bdata["99.0"][1]
            qcvars["HF TOTAL ENERGY"] = ehf
            qcvars["SCF TOTAL ENERGY"] = ehf
            qcvars["CURRENT REFERENCE ENERGY"] = ehf
            qcvars["MP2 CORRELATION ENERGY"] = emp2 - ehf
            qcvars["CURRENT CORRELATION ENERGY"] = emp2 - ehf
            properties["mp2_correlation_energy"] = emp2 - ehf

        # Correct CCSD because its odd?
        # if input_model.model.method.lower() == "ccsd":
        #     m1 = re.findall(" CCSD correlation energy.+=.+\d+\.\d+", outfiles["dispatch.out"])
        #     m2 = re.findall(" CCSD total energy.+=.+\d+\.\d+", outfiles["dispatch.out"])

        props, prov = self._parse_logfile_common(outtext, input_model.dict())
        output_data["provenance"] = prov
        output_data["properties"] = properties
        output_data["properties"].update(props)
        output_data["stdout"] = outfiles["dispatch.out"]
        output_data["success"] = True

        merged_data = {**input_model.dict(), **output_data}
        merged_data["extras"]["qcvars"] = qcvars

        return AtomicResult(**merged_data)

    def _parse_logfile_common(self, outtext: str, input_dict: Dict[str, Any]):
        """
        Parses fields from log file that are not parsed from QCSCRATCH in parse_output
        """

        properties = {}
        provenance = Provenance(creator="QChem", version=self.get_version(), routine="qchem").dict()
        mobj = re.search(r"This is a multi-thread run using ([0-9]+) threads", outtext)
        if mobj:
            provenance["nthreads"] = int(mobj.group(1))

        mobj = re.search(r"Total job time:\s*" + NUMBER + r"s\(wall\)", outtext)
        if mobj:
            provenance["wall_time"] = float(mobj.group(1))

        mobj = re.search(r"Archival summary:\s*\n[0-9]+\\[0-9+]\\([\w\.\-]+)\\", outtext)
        if mobj:
            provenance["hostname"] = mobj.group(1)

        mobj = re.search(r"\n\s*There are\s+(\d+) alpha and\s+(\d+) beta electrons\s*\n", outtext)
        if mobj:
            properties["calcinfo_nalpha"] = int(mobj.group(1))
            properties["calcinfo_nbeta"] = int(mobj.group(2))

        mobj = re.search(r"\n\s*There are\s+\d+ shells and\s+(\d+) basis functions\s*\n", outtext)
        if mobj:
            properties["calcinfo_nbasis"] = int(mobj.group(1))

        mobj = re.search(r"\n\s*RI-MP2 CORRELATION ENERGY\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        if mobj:
            properties["mp2_correlation_energy"] = float(mobj.group(1))

        mobj = re.search(r"\n\s*RI-MP2 SINGLES ENERGY\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        if mobj:
            properties["mp2_singles_energy"] = float(mobj.group(1))

        mobj_aaaa = re.search(r"\n\s*RI-MP2 ENERGY \(aa\|aa\)\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        mobj_bbbb = re.search(r"\n\s*RI-MP2 ENERGY \(bb\|bb\)\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        if mobj_aaaa and mobj_bbbb:
            properties["mp2_same_spin_correlation_energy"] = float(mobj_aaaa.group(1)) + float(mobj_bbbb.group(1))

        mobj_aabb = re.search(r"\n\s*RI-MP2 ENERGY \(aa\|bb\)\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        mobj_bbaa = re.search(r"\n\s*RI-MP2 ENERGY \(bb\|aa\)\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        if mobj_aaaa and mobj_bbbb:
            properties["mp2_opposite_spin_correlation_energy"] = float(mobj_aabb.group(1)) + float(mobj_bbaa.group(1))

        properties["calcinfo_natom"] = len(input_dict["molecule"]["symbols"])

        mobj = re.search(r"\n\s*(\d+)\s+" + NUMBER + "\s+" + NUMBER + r"\s+Convergence criterion met\s*\n", outtext)
        if mobj:
            properties["scf_iterations"] = int(mobj.group(1))

        mobj = re.search(
            r"\n\s+Dipole Moment \(Debye\)\s*\n\s+X\s+" + NUMBER + r"\s+Y\s+" + NUMBER + r"\s+Z\s+" + NUMBER + r"\s*\n",
            outtext,
        )
        if mobj:
            cf = constants.conversion_factor("debye", "e * bohr")
            properties["scf_dipole_moment"] = [float(mobj.group(i)) * cf for i in range(1, 4)]

        return properties, provenance

    def parse_logfile(self, outfiles: Dict[str, str]) -> AtomicResult:
        """
        Parses a log file.
        """
        warnings.warn(
            "parse_logfile will result in precision loss for some fields due to trunctation in " "Q-Chem output files."
        )

        outtext = outfiles["dispatch.out"]

        mobj = re.search(r"(?:User input\:|Running Job?)\s+\d+\s+of\s+(\d+)", outtext)
        if mobj:
            if int(mobj.group(1)) > 1:
                raise InputError("Multi-job Q-Chem log files not supported.")

        input_dict = {}
        mobj = re.search(r"\n-{20,}\nUser input:\n-{20,}\n(.+)\n-{20,}", outtext, re.DOTALL)
        if mobj:
            inputtext = mobj.group(1)

            rem_match = re.search(r"\$rem\s*\n([^\$]+)\n\s*\$end", inputtext, re.DOTALL | re.IGNORECASE)
            if rem_match:
                rem_text = rem_match.group(1)
                lines = rem_text.split("\n")
                keywords = {}
                for line in lines:
                    s = re.sub(r"(^|[^\\])!.*", "", line).split()
                    if len(s) == 0:
                        continue
                    keywords[s[0].lower()] = s[1].lower()
                input_dict["model"] = {}
                input_dict["model"]["method"] = keywords.pop("method").lower()
                input_dict["model"]["basis"] = keywords.pop("basis").lower()
                if "jobtype" in keywords:
                    jobtype = keywords.pop("jobtype")
                else:
                    jobtype = "sp"
                _jobtype_to_driver = {
                    "sp": "energy",
                    "force": "gradient",
                    "freq": "hessian",
                }  # optimization intentionally not supported
                try:
                    input_dict["driver"] = _jobtype_to_driver[jobtype]
                except KeyError:
                    raise KeyError(f"Jobtype {jobtype} not supported in qchem log file parser.")

                for key in keywords:
                    if keywords[key] == "false":
                        keywords[key] = False
                    if keywords[key] == "true":
                        keywords[key] = True
                input_dict["keywords"] = keywords

            molecule_match = re.search(r"\$molecule\s*\n([^\$]+)\n\s*\$end", inputtext, re.DOTALL | re.IGNORECASE)
            if molecule_match:
                molecule_text = molecule_match.group(1)
                if keywords.get("input_bohr", False):
                    molecule_text += "\nunits au"
                molecule = Molecule.from_data(molecule_text, dtype="psi4")
                input_dict["molecule"] = molecule.dict()

            _excluded_rem = {
                "input_bohr",
                "mem_total",
                "mem_static",
            }  # parts of the rem normally written by build_input, and which do not affect results
            for item in _excluded_rem:
                if item in input_dict["keywords"]:
                    input_dict["keywords"].pop(item)

        try:
            qcscr_result = self.parse_output(outfiles, AtomicInput(**input_dict)).dict()
        except KeyError:
            props, prov = self._parse_logfile_common(outtext, input_dict)
            qcscr_result = {"properties": props, "provenance": prov, **input_dict}

        mobj = re.search(r"\n\s*Total\s+energy in the final basis set =\s+" + NUMBER + r"\s*\n", outtext)
        if mobj and qcscr_result["properties"].get("scf_total_energy", None) is None:
            qcscr_result["properties"]["scf_total_energy"] = float(mobj.group(1))

        mobj = re.search(r"\n\s*Nuclear Repulsion Energy =\s+" + NUMBER + r"\s+hartrees\s*\n", outtext)
        if mobj and qcscr_result["properties"].get("nuclear_repulsion_energy", None) is None:
            qcscr_result["properties"]["nuclear_repulsion_energy"] = float(mobj.group(1))

        mobj = re.search(r"\n\s*RI-MP2 TOTAL ENERGY\s+=\s+" + NUMBER + r"\s+au\s*\n", outtext)
        if mobj and qcscr_result["properties"].get("mp2_total_energy", None) is None:
            qcscr_result["properties"]["mp2_total_energy"] = float(mobj.group(1))

        _scf_methods = {
            "hf",
            "spw92",
            "lda",
            "svwn5",
            "b97-d3(0)",
            "b97-d",
            "pbe",
            "blyp",
            "revpbe",
            "beef-vdw",
            "bop",
            "bp86",
            "bp86vwn",
            "bpbe",
            "edf1",
            "edf2",
            "gam",
            "hcth93",
            "hcth120",
            "hcth147",
            "hcth407",
            "hle16",
            "kt1",
            "kt2",
            "kt3",
            "mpw91",
            "n12",
            "olyp",
            "pbeop",
            "pbesol",
            "pw91",
            "rpbe",
            "rvv10",
            "sogga",
            "sogga11",
            "vv10",
            "b97m-v",
            "b97m-rv",
            "m06-l",
            "tpss",
            "revtpss",
            "bloc",
            "m11-l",
            "mbeef",
            "mgga_ms0",
            "mgga_ms1",
            "mgga_ms2",
            "mgga_mvs",
            "mn12-l",
            "mn15-l",
            "otpss",
            "pkzb",
            "scan",
            "t-hcth",
            "tm",
            "vsxc",
            "b3lyp",
            "pbe0",
            "revpbe0",
            "b97",
            "b1lyp",
            "b1pw91",
            "b3lyp5",
            "b3p86",
            "b1lyp",
            "b1pw91",
            "b3lyp5",
            "b3p86",
            "b3pw91",
            "b5050lyp",
            "b97-1",
            "b97-2",
            "b97-3",
            "b97-k",
            "bhhlyp",
            "hflyp",
            "mpw1k",
            "mpw1lyp",
            "mpw1pbe",
            "mpw1pw91",
            "o3lyp",
            "pbeh-3c",
            "pbe50",
            "sogga11-x",
            "wc04",
            "wp04",
            "x3lyp",
            "m06-2x",
            "m08-hx",
            "tpssh",
            "revtpssh",
            "b1b95",
            "b3tlap",
            "bb1k",
            "bmk",
            "dldf",
            "m05",
            "m05-2x",
            "m06",
            "m06-hf",
            "m08-so",
            "mgga_ms2h",
            "mgga_mvsh",
            "mn15",
            "mpw1b95",
            "mpwb1k",
            "pw6b95",
            "pwb6k",
            "scan0",
            "t-hcthh",
            "tpss0",
            "wb97x-v",
            "wb97x-d3",
            "wb97x-d",
            "cam-b3lyp",
            "cam-qtp00",
            "cam-qtp01",
            "hse-hjs",
            "lc-rvv10",
            "lc-vv10",
            "lc-wpbe08",
            "lrc-bop",
            "lrc-wpbe",
            "lrc-wpbeh",
            "n12-sx",
            "rcam-b3lyp",
            "wb97",
            "wb97x",
            "wb97x-rv",
            "wb97m-v",
            "m11",
            "mn12-sx",
            "wb97m-rv",
            "wm05-d",
            "wm06-d3",
            "dsd-pbepbe-d3",
            "wb97x-2(lp)",
            "wb97x-2(tqz)",
            "xyg3",
            "xygj-os",
            "b2plyp",
            "b2gpplyp",
            "dsd-pbep86-d3",
            "ls1dh-pbe",
            "pbe-qidh",
            "pbe0-2",
            "pbe0-dh",
            "ptpss-d3",
            "dsd-pbeb95-d3",
            "pwpb95-d3",
        }
        _mp2_methods = {"rimp2"}

        method = input_dict["model"]["method"].lower()
        if qcscr_result["properties"].get("return_energy", None) is None:
            if method in _mp2_methods:
                qcscr_result["properties"]["return_energy"] = qcscr_result["properties"]["mp2_total_energy"]
            elif method in _scf_methods:
                qcscr_result["properties"]["return_energy"] = qcscr_result["properties"]["scf_total_energy"]
            else:
                raise NotImplementedError(f"Method {method} not supported by logfile parser for energy driver.")

        if input_dict["driver"] == "gradient" and qcscr_result.get("return_result", None) is None:

            def read_matrix(text):
                lines = text.split("\n")
                i = 0
                mdict = defaultdict(dict)
                maxcol = 0
                maxrow = 0
                while i < len(lines):
                    cols = [int(idx) for idx in lines[i].split()]
                    maxcol = max(maxcol, *cols)
                    i += 1
                    while i < len(lines):
                        s = lines[i].split()
                        if len(s) <= len(cols):
                            break
                        assert len(s) == len(cols) + 1, s
                        row = int(s[0])
                        maxrow = max(maxrow, row)
                        data = [float(field) for field in s[1:]]
                        for col_idx, col in enumerate(cols):
                            mdict[row - 1][col - 1] = data[col_idx]
                        i += 1

                ret = np.zeros((maxrow, maxcol))
                for row in mdict:
                    for col in mdict[row]:
                        ret[row, col] = mdict[row][col]
                return ret

            if method in _mp2_methods:
                mobj = re.search(
                    r"\n\s+Full Analytical Gradient of MP2 Energy \(in au.\)\s*\n"
                    r"([\s\d\.EDed\+\-]+)\n"
                    r"\s*Gradient time:",
                    outtext,
                )

            elif method in _scf_methods:
                mobj = re.search(
                    r"\n\s+Gradient of SCF Energy\s*\n([\s\d\.EDed\+\-]+)\n\s*Max gradient component =", outtext
                )

            else:
                raise NotImplementedError(f"Method {method} not supported by the logfile parser for gradient driver.")

            if mobj:
                qcscr_result["return_result"] = read_matrix(mobj.group(1)).T

        qcscr_result["success"] = True  # XXX: have a nice day?
        qcscr_result["stdout"] = outtext
        if input_dict["driver"] == "energy" and qcscr_result.get("return_result", None) is None:
            qcscr_result["return_result"] = qcscr_result["properties"]["return_energy"]

        return AtomicResult(**qcscr_result)



"""
Imports the various procedure backends
"""

from typing import Set

from ..exceptions import InputError, ResourceError
from .berny import BernyProcedure
from .geometric import GeometricProcedure
from .nwchem_opt import NWChemDriverProcedure
from .optking import OptKingProcedure
from .torsiondrive import TorsionDriveProcedure
from .model import ProcedureHarness

__all__ = ["register_procedure", "get_procedure", "list_all_procedures", "list_available_procedures"]

procedures = {}


def register_procedure(entry_point: ProcedureHarness) -> None:
    """
    Register a new ProcedureHarness with QCEngine
    """

    name = entry_point.name
    if name.lower() in procedures.keys():
        raise ValueError("{} is already a registered procedure.".format(name))

    procedures[name.lower()] = entry_point


def get_procedure(name: str) -> ProcedureHarness:
    """
    Returns a procedures executor class
    """

    name = name.lower()

    if name not in procedures:
        raise InputError(f"Procedure {name} is not registered to QCEngine.")

    ret = procedures[name]
    if not ret.found():
        raise ResourceError(f"Procedure {name} is registered with QCEngine, but cannot be found.")

    return ret


def list_all_procedures() -> Set[str]:
    """
    List all procedures registered by QCEngine.
    """
    return set(procedures.keys())


def list_available_procedures() -> Set[str]:
    """
    List all procedures that can be exectued (found) by QCEngine.
    """

    ret = set()
    for k, p in procedures.items():
        if p.found():
            ret.add(k)

    return ret


register_procedure(GeometricProcedure())
register_procedure(OptKingProcedure())
register_procedure(BernyProcedure())
register_procedure(NWChemDriverProcedure())
register_procedure(TorsionDriveProcedure())



"""Harness for TeraChem Frontend"""
import logging
from os import getenv
from typing import Any, Dict

from .terachem_pbs import TeraChemPBSHarness, _pbs_defaults

logger = logging.getLogger(__name__)

_fe_defaults = {
    "name": "terachem_fe",
}


class TeraChemFrontEndHarness(TeraChemPBSHarness):
    """QCEngine Harness for interfacing with the TeraChem Frontend (Protocol Buffer Server + file server)"""

    _defaults = {**_pbs_defaults, **_fe_defaults}
    _tcpb_min_version = "0.9.0"
    _tcpb_client = "TCFrontEndClient"
    _env_vars: Dict[str, Any] = {
        **TeraChemPBSHarness._env_vars,
        **{
            "frontend_host": getenv("TERACHEM_FE_HOST", "127.0.0.1"),
            "frontend_port": int(getenv("TERACHEM_FE_PORT", 80)),
            "uploads_prefix": getenv("TERACHEM_FE_UPLOADS_PREFIX", "uploads"),
        },
    }
    _env_vars_external: str = (
        TeraChemPBSHarness._env_vars_external + " and/or TERACHEM_FE_HOST, TERACHEM_FE_PORT, TERACHEM_FE_UPLOADS_PREFIX"
    )



import logging
import re
from decimal import Decimal

import numpy as np
import qcelemental as qcel
from qcelemental.models import Molecule
from qcelemental.molparse import regex

from ..util import PreservingDict, load_hessian

logger = logging.getLogger(__name__)


def harvest_output(outtext):
    """Function to separate portions of a CFOUR output file *outtest*,
    divided by xjoda.

    """
    pass_psivar = []
    pass_coord = []
    pass_grad = []

    # for outpass in re.split(r'--invoking executable xjoda', outtext, re.MULTILINE):
    for outpass in re.split(r"JODA beginning optimization cycle", outtext, re.MULTILINE):
        psivar, c4coord, c4grad, version, module, error = harvest_outfile_pass(outpass)
        pass_psivar.append(psivar)
        pass_coord.append(c4coord)
        pass_grad.append(c4grad)

        # print('\n\nXXXXXXXXXXXXXXXXXXXXXXXXXX\n\n')
        # print(outpass)
        # print(psivar, c4coord, c4grad, version, error)
        # print('\n\nxxxxxxxxxxxxxxxxxxxxxxxxxx\n\n')

    retindx = -1 if pass_coord[-1] else -2

    #    print '    <<<  C4 PSIVAR  >>>'
    #    for item in pass_psivar[retindx]:
    #        print('       %30s %16.8f' % (item, pass_psivar[retindx][item]))
    #    print '    <<<  C4 COORD   >>>'
    #    for item in pass_coord[retindx]:
    #        print('       %16.8f %16.8f %16.8f' % (item[0], item[1], item[2]))
    #    print '    <<<   C4 GRAD   >>>'
    #    for item in pass_grad[retindx]:
    #        print('       %16.8f %16.8f %16.8f' % (item[0], item[1], item[2]))

    return pass_psivar[retindx], pass_coord[retindx], pass_grad[retindx], version, module, error


def harvest_outfile_pass(outtext):
    """Function to read CFOUR output file *outtext* and parse important
    quantum chemical information from it in

    """
    psivar = PreservingDict()
    psivar_coord = None
    psivar_grad = None
    version = ""
    module = None
    error = ""

    #    TODO: CI
    #          other ROHF tests

    NUMBER = r"(?x:" + regex.NUMBER + ")"
    DECIMAL = r"(?x:" + regex.DECIMAL + ")"

    # Process version
    mobj = re.search(r"^\s*" + r"Version" + r"\s+" + r"(?P<version>[\w.]+)" + r"\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("matched version")
        version = mobj.group("version")

    # Process NRE
    mobj = re.search(
        r"^\s+" + r"(?:Nuclear repulsion energy :)" + r"\s+" + NUMBER + r"\s+a\.u\.\s*$",
        outtext,
        re.MULTILINE | re.IGNORECASE,
    )
    if mobj:
        logger.debug("matched nre")
        psivar["NUCLEAR REPULSION ENERGY"] = mobj.group(1)

    # Process calcinfo
    mobj = re.search(
        r"^\s*" + r"There are" + r"\s+" + r"(?P<nbf>\d+)" + r"\s+" + r"functions in the AO basis." + r"\s*$",
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched nbf", mobj.groups())
        psivar["N BASIS FUNCTIONS"] = mobj.group("nbf")
        psivar["N MOLECULAR ORBITALS"] = mobj.group("nbf")  # TODO BAD

    mobj = re.search(
        # fmt: off
        r"^\s*" + "Alpha population by irrep:" + r"(?P<aocc>[\d\s]+)" + r"\s*" +
        r"^\s*" + "Beta population by irrep:" + r"(?P<bocc>[\d\s]+)" + r"\s*",
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched occupied", mobj.groups())
        psivar["N ALPHA ELECTRONS"] = sum([int(d) for d in mobj.group("aocc").split()])
        psivar["N BETA ELECTRONS"] = sum([int(d) for d in mobj.group("bocc").split()])

    # Process SCF
    mobj = re.search(r"^\s+" + r"(?:E\(SCF\))" + r"\s+=\s+" + NUMBER + r"\s+a\.u\.\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("matched scf1")
        psivar["SCF TOTAL ENERGY"] = mobj.group(1)

    mobj = re.search(r"^\s+" + r"(?:E\(SCF\)=)" + r"\s+" + NUMBER + r"\s+" + NUMBER + r"\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("matched scf2")
        mobj5 = re.search(r"BRUECKNER\s+IBRKNR\s+ON", outtext)
        if not mobj5:
            # interferes with ROHF BCCD
            psivar["SCF TOTAL ENERGY"] = mobj.group(1)

    if "SCF TOTAL ENERGY" not in psivar:
        # can be too greedy and match across scf cycles
        mobj = re.search(
            # fmt: off
            r'^\s+' + r'(?:SCF has converged.)' + r'\s*$' +
            r'(?:.*?)' +
            r'^\s+' + r'(?:\d+)' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj:
            logger.debug("matched scf3")
            psivar["SCF TOTAL ENERGY"] = mobj.group(1)

    mobj = re.search(r"^\s+" + r"(?:E\(ROHF\)=)" + r"\s+" + NUMBER + r"\s+" + NUMBER + r"\s*$", outtext, re.MULTILINE)
    if mobj:
        logger.debug("matched scf4")
        psivar["SCF TOTAL ENERGY"] = mobj.group(1)

    # Process MP2
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:E2\(AA\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(AB\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(TOT\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:Total MP2 energy)' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*$',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched mp2r")
        psivar["MP2 SAME-SPIN CORRELATION ENERGY"] = 2 * Decimal(mobj.group(1))
        psivar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(2)
        psivar["MP2 CORRELATION ENERGY"] = 2 * Decimal(mobj.group(1)) + Decimal(mobj.group(2))
        psivar["MP2 TOTAL ENERGY"] = mobj.group(4)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:E2\(AA\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(BB\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(AB\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(TOT\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:Total MP2 energy)' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*$',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched mp2u")
        psivar["MP2 SAME-SPIN CORRELATION ENERGY"] = Decimal(mobj.group(1)) + Decimal(mobj.group(2))
        psivar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(3)
        psivar["MP2 CORRELATION ENERGY"] = Decimal(mobj.group(1)) + Decimal(mobj.group(2)) + Decimal(mobj.group(3))
        psivar["MP2 TOTAL ENERGY"] = mobj.group(5)

    mobj = re.search(
        # particularly, want to avoid capture when following line present:
        #  "MP2 energies are correct only for semicanonical orbitals."
        # fmt: off
        r'Singles contribution will be calculated.' + r'\s*' +
        r'^\s+' + r'-*' + r'\s*' +
        r'^\s+' + r'(?:E\(SCF\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(AA\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(BB\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(AB\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(SINGLE\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:E2\(TOT\))' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'(?:Total MP2 energy)' + r'\s+=\s+' + NUMBER + r'\s+a.u.\s*$',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched mp2ro")
        psivar["MP2 SAME-SPIN CORRELATION ENERGY"] = Decimal(mobj.group(2)) + Decimal(mobj.group(3))
        psivar["MP2 OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(4)
        psivar["MP2 SINGLES ENERGY"] = mobj.group(5)
        psivar["MP2 CORRELATION ENERGY"] = (
            Decimal(mobj.group(5)) + Decimal(mobj.group(2)) + Decimal(mobj.group(3)) + Decimal(mobj.group(4))
        )
        psivar["MP2 TOTAL ENERGY"] = mobj.group(7)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:S-MBPT\(2\))' + r'\s+' + r'(?P<sgl>' + NUMBER + r')' + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(2\))' + r'\s+' + r'(?P<dbl>' + NUMBER + r')' + r'\s+' +
                                                r'(?P<mp2tot>' + NUMBER + r')' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched mp2ro2")
        # psivar['MP2 SAME-SPIN CORRELATION ENERGY'] = Decimal(mobj.group(1)) + Decimal(mobj.group(2))
        # psivar['MP2 OPPOSITE-SPIN CORRELATION ENERGY'] = mobj.group(3)
        psivar["MP2 SINGLES ENERGY"] = mobj.group("sgl")
        psivar["MP2 CORRELATION ENERGY"] = Decimal(mobj.group("sgl")) + Decimal(mobj.group("dbl"))
        psivar["MP2 TOTAL ENERGY"] = mobj.group("mp2tot")

    # Process MP3
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:D-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp3r")
        dmp2 = Decimal(mobj.group(1))
        dmp3 = Decimal(mobj.group(3))
        psivar["MP2 CORRELATION ENERGY"] = dmp2
        psivar["MP2 TOTAL ENERGY"] = mobj.group(2)
        psivar["MP3 CORRELATION ENERGY"] = dmp2 + dmp3
        psivar["MP3 TOTAL ENERGY"] = mobj.group(4)
        psivar["MP2.5 CORRELATION ENERGY"] = dmp2 + Decimal("0.500000000000") * dmp3
        psivar["MP2.5 TOTAL ENERGY"] = psivar["MP2.5 CORRELATION ENERGY"] + psivar["SCF TOTAL ENERGY"]
        psivar["MP3 SINGLES ENERGY"] = Decimal("0.0")
        psivar["MP3 DOUBLES ENERGY"] = dmp2 + dmp3
        module = "vcc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:S-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:S-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp3ro")
        dmp2 = Decimal(mobj.group(1)) + Decimal(mobj.group(3))
        dmp3 = Decimal(mobj.group(5)) + Decimal(mobj.group(7))
        psivar["MP3 CORRELATION ENERGY"] = dmp2 + dmp3
        psivar["MP3 TOTAL ENERGY"] = mobj.group(8)
        psivar["MP2.5 CORRELATION ENERGY"] = dmp2 + Decimal("0.500000000000") * dmp3
        psivar["MP2.5 TOTAL ENERGY"] = psivar["MP2.5 CORRELATION ENERGY"] + psivar["SCF TOTAL ENERGY"]
        psivar["MP3 SINGLES ENERGY"] = Decimal(mobj.group(1)) + Decimal(mobj.group(5))
        psivar["MP3 DOUBLES ENERGY"] = Decimal(mobj.group(3)) + Decimal(mobj.group(7))
        module = "vcc"

    mobj = re.search(
        # fmt: off
        r"^\s*" + r"(?:MP2 correlation energy:)\s+" + r"(?P<mp2corl>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:MP3 correction:)\s+" +         r"(?P<mp3corr>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:MP3 correlation energy:)\s+" + r"(?P<mp3corl>" + NUMBER + ")" + r"\s*" +
        r"(?:.*?)" +
        r"^\s*" + r"(?:Non-iterative calculation of MP3)" + r".*" +
        r"(?:.*?)" +
        r"^\s*" + r"(?:Total MP3 energy:)" + r"\s+" + r"(?P<mp3tot>" + NUMBER + ")" + r"\s*$",
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp3 ncc")
        # psivar["MP2 CORRELATION ENERGY"] = mobj.group("mp2corl")
        psivar["MP3 CORRELATION ENERGY"] = mobj.group("mp3corl")
        psivar["MP3 CORRECTION ENERGY"] = mobj.group("mp3corr")
        psivar["MP3 TOTAL ENERGY"] = mobj.group("mp3tot")
        # looks like ncc is rhf-only
        # psivar["MP2 DOUBLES ENERGY"] = mobj.group("mp2corl")
        psivar["MP3 DOUBLES ENERGY"] = mobj.group("mp3corl")
        module = "ncc"

    # Process MP4
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:D-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:Q-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:S-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp4r")
        dmp2 = Decimal(mobj.group(1))
        dmp3 = Decimal(mobj.group(3))
        dmp4sdq = Decimal(mobj.group(5)) + Decimal(mobj.group(7)) + Decimal(mobj.group(9))
        psivar["MP2 CORRELATION ENERGY"] = dmp2
        psivar["MP2 TOTAL ENERGY"] = mobj.group(2)
        psivar["MP3 CORRELATION ENERGY"] = dmp2 + dmp3
        psivar["MP3 TOTAL ENERGY"] = mobj.group(4)
        psivar["MP2.5 CORRELATION ENERGY"] = dmp2 + Decimal("0.500000000000") * dmp3
        psivar["MP2.5 TOTAL ENERGY"] = psivar["MP2.5 CORRELATION ENERGY"] + psivar["SCF TOTAL ENERGY"]
        psivar["MP4(SDQ) CORRELATION ENERGY"] = dmp2 + dmp3 + dmp4sdq
        psivar["MP4(SDQ) TOTAL ENERGY"] = mobj.group(10)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:S-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(2\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:S-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:D-MBPT\(3\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:L-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:NL-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp4ro")
        module = "vcc"
        dmp2 = Decimal(mobj.group(1)) + Decimal(mobj.group(3))
        dmp3 = Decimal(mobj.group(5)) + Decimal(mobj.group(7))
        dmp4sdq = Decimal(mobj.group(9)) + Decimal(mobj.group(11))
        psivar["MP2 CORRELATION ENERGY"] = dmp2
        psivar["MP2 TOTAL ENERGY"] = mobj.group(4)
        psivar["MP3 CORRELATION ENERGY"] = dmp2 + dmp3
        psivar["MP3 TOTAL ENERGY"] = mobj.group(8)
        psivar["MP2.5 CORRELATION ENERGY"] = dmp2 + Decimal("0.500000000000") * dmp3
        psivar["MP2.5 TOTAL ENERGY"] = psivar["MP2.5 CORRELATION ENERGY"] + psivar["SCF TOTAL ENERGY"]
        psivar["MP4(SDQ) CORRELATION ENERGY"] = dmp2 + dmp3 + dmp4sdq
        psivar["MP4(SDQ) TOTAL ENERGY"] = mobj.group(12)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:D-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:Q-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:S-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:T-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp4tr")
        dmp4sdq = Decimal(mobj.group(1)) + Decimal(mobj.group(3)) + Decimal(mobj.group(5))
        dmp4t = Decimal(mobj.group(7))
        psivar["MP4(SDQ) CORRELATION ENERGY"] = psivar["MP3 CORRELATION ENERGY"] + dmp4sdq
        psivar["MP4(SDQ) TOTAL ENERGY"] = mobj.group(6)
        psivar["MP4(T) CORRECTION ENERGY"] = dmp4t
        psivar["MP4 CORRELATION ENERGY"] = psivar["MP3 CORRELATION ENERGY"] + dmp4sdq + dmp4t
        psivar["MP4 TOTAL ENERGY"] = mobj.group(8)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:L-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:NL-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:WT12-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*' +
        r'^\s+' + r'(?:T-MBPT\(4\))' + r'\s+' + NUMBER + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp4tro")
        module = "vcc"
        dmp4sdq = Decimal(mobj.group(1)) + Decimal(mobj.group(3))
        dmp4t = Decimal(mobj.group(5)) + Decimal(mobj.group(7))  # WT12 with T, not SDQ
        psivar["MP4(SDQ) CORRELATION ENERGY"] = psivar["MP3 CORRELATION ENERGY"] + dmp4sdq
        psivar["MP4(SDQ) TOTAL ENERGY"] = mobj.group(4)
        psivar["MP4(T) CORRECTION ENERGY"] = dmp4t
        psivar["MP4 CORRELATION ENERGY"] = psivar["MP3 CORRELATION ENERGY"] + dmp4sdq + dmp4t
        psivar["MP4 TOTAL ENERGY"] = mobj.group(8)

    mobj = re.search(
        # fmt: off
        r"^\s*" + r"(?:MP2 correlation energy:)\s+" + r"(?P<mp2corl>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:MP3 correction:)\s+" +         r"(?P<mp3corr>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:MP3 correlation energy:)\s+" + r"(?P<mp3corl>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:SDQ-MP4 correction:)\s+" + r"(?P<mp4sdqcorr>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:SDQ-MP4 correlation energy:)\s+" + r"(?P<mp4sdqcorl>" + NUMBER + ")" + r"\s*" +
        r"(" +
        r"^\s*" + r"(?:T-MP4 correction:)\s+" + r"(?P<mp4tcorr>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:Total MP4 correction:)\s+" + r"(?P<mp4corr>" + NUMBER + ")" + r"\s*" +
        r"^\s*" + r"(?:MP4 correlation energy:)\s+" + r"(?P<mp4sdtqcorl>" + NUMBER + ")" + r"\s*" +
        r")?" +
        r"(?:.*?)" +
        r"^\s*" + r"(?:Non-iterative calculation of (MP4|SDQ-MP4))" + r".*" +
        r"(?:.*?)" +
        r"^\s*" + r"(?:Total (?P<mp4flavor>(MP4|SDQ-MP4)) energy:)" + r"\s+" + r"(?P<mp4flavortot>" + NUMBER + ")" + r"\s*$",
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched mp4 ncc")
        # psivar["MP2 CORRELATION ENERGY"] = mobj.group("mp2corl")
        module = "ncc"
        mtd = {"MP4": "MP4", "SDQ-MP4": "MP4(SDQ)"}[mobj.group("mp4flavor")]
        psivar["MP3 CORRELATION ENERGY"] = mobj.group("mp3corl")
        psivar["MP3 CORRECTION ENERGY"] = mobj.group("mp3corr")
        psivar["MP4(SDQ) CORRELATION ENERGY"] = mobj.group("mp4sdqcorl")
        # looks like ncc is rhf-only
        # psivar["MP2 DOUBLES ENERGY"] = mobj.group("mp2corl")
        psivar["MP3 DOUBLES ENERGY"] = mobj.group("mp3corl")
        if mtd == "MP4(SDQ)":
            psivar["MP4(SDQ) TOTAL ENERGY"] = mobj.group("mp4flavortot")
        elif mtd == "MP4":
            psivar["MP4(T) CORRECTION ENERGY"] = mobj.group("mp4tcorr")
            psivar["MP4 CORRECTION ENERGY"] = mobj.group("mp4corr")
            psivar["MP4 TOTAL ENERGY"] = mobj.group("mp4flavortot")
            psivar["MP4 CORRELATION ENERGY"] = mobj.group("mp4sdtqcorl")

    # Process CI Iterations
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?P<fullCI>(?P<iterCI>Q?CI(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:\d+)' + r'\s+' + r"(?P<corl>" + NUMBER + r")" + r'\s+' + r"(?P<tot>" + NUMBER + r")" + r'\s+DIIS\s*' +
        r'^\s*(?:-+)\s*' +
        # CI iterations for CISD, CC iterations for QCISD
        r'^\s*(?:A miracle (?P<ccprog>has come|come) to pass. The (CI|CC) iterations have converged.)\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ci with full %s iterating %s" % (mobj.group("fullCI"), mobj.group("iterCI")))
        module = {"has come": "vcc", "come": "ecc"}[mobj.group("ccprog")]

        mtd = mobj.group("iterCI").upper()
        psivar[f"{mtd} CORRELATION ENERGY"] = mobj.group("corl")
        psivar[f"{mtd} TOTAL ENERGY"] = mobj.group("tot")

        mobj2 = re.search(
            # fmt: off
            r"^\s+" + r"E\(QCISD\)\s+=\s+" + r"(?P<qcisd>" + NUMBER + r")" + r"\s*" +
            r"^\s+" + r"E\(QCISD\(T\)\)\s+=\s+" + r"(?P<qcisdt>" + NUMBER + r")" + r"\s*$",
            # fmt: on
            outtext,
            re.MULTILINE | re.DOTALL,
        )
        if mobj2 and mobj.group("fullCI") == "QCISD(T)":
            logger.debug("matched qcisd(t)")
            psivar["QCISD(T) TOTAL ENERGY"] = mobj2.group("qcisdt")
            psivar["QCISD(T) CORRECTION ENERGY"] = Decimal(mobj2.group("qcisdt")) - Decimal(mobj2.group("qcisd"))
            psivar["QCISD(T) CORRELATION ENERGY"] = psivar["QCISD(T) TOTAL ENERGY"] - psivar["SCF TOTAL ENERGY"]

    # Process CC Iterations
    cciterations = re.compile(
        # fmt: off
        r'^\s+' + r'(?P<fullCC>(?P<iterCC>L?CC(?:\w+(?:-(?:1|1b|2|3))?))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:\d+)' + r'\s+' + r"(?P<corl>" + NUMBER + r")" + r'\s+' + r"(?P<tot>" + NUMBER + r")" + r'\s+DIIS\s*' +
        r'^\s*(?:-+)\s*' +
        r'^\s*(?:A miracle (?P<ccprog>has come|come) to pass. The CC iterations have converged.)\s*$',
        # fmt: on
        re.MULTILINE
        | re.DOTALL,
    )
    mobj_list = [m.groupdict() for m in cciterations.finditer(outtext)]
    if mobj_list:
        first_match = mobj_list[0]
        last_match = mobj_list[-1]

        logger.debug("matched cc with full %s iterating %s" % (last_match["fullCC"], last_match["iterCC"]))
        module = {"has come": "vcc", "come": "ecc"}[last_match["ccprog"]]

        mobj4 = re.search(r"CALCLEVEL\s+ICLLVL\s+CCSDT-1b", outtext)
        mtd = last_match["iterCC"].upper()
        if mtd == "CCSDT-1":
            if mobj4 and module == "vcc":
                mtd = "CCSDT-1B"
            else:
                mtd = "CCSDT-1A"
        elif mtd == "CCSDT-1b":
            mtd = "CCSDT-1B"

        mobj5 = re.search(r"BRUECKNER\s+IBRKNR\s+ON", outtext)
        if mobj5 and mtd == "CCSD":
            # BCCD corl = last Brueckner iter total E - first Brueckner iter HF E
            psivar[f"BCCD CORRELATION ENERGY"] = Decimal(last_match["tot"]) - psivar["SCF TOTAL ENERGY"]
            psivar[f"BCCD TOTAL ENERGY"] = last_match["tot"]
            logger.debug("matched Brueckner iter")
        else:
            psivar[f"{mtd} CORRELATION ENERGY"] = last_match["corl"]
            psivar[f"{mtd} TOTAL ENERGY"] = last_match["tot"]

            mobj3 = re.search(r"SCF reference function:  RHF", outtext)
            if mobj3 and mtd not in ["CCSDT-1A", "CCSDT-1B", "CCSDT-2", "CCSDT-3", "CCSDT"]:
                psivar[f"{mtd} DOUBLES ENERGY"] = last_match["corl"]

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:\d+)' + r'\s+' + r'(?P<corl>' + NUMBER + r')\s+' +
                  NUMBER + r'\s+' + NUMBER + r'\s+' +
                  NUMBER + r'\s+' + NUMBER +  r'\s+' +
                  r'(' + NUMBER +  r')?' + r'(' + r'\s+' + NUMBER + r')?' + r'\s*' +
        r'^\s*' +
        r'^\s*' + r'(?:\w+(?:-(1a|1b|2|3))? iterations converged .*?)' +
        r'^\s*' +
        r'^\s*' + r'(?:Total (?P<iterCC>\w+(?:-(1a|1b|2|3))?) energy:)' + r'\s+' + r'(?P<tot>' + NUMBER + r')\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ncc cc iter")
        # looks like ncc is rhf-only
        mtd = mobj.group("iterCC").upper()
        psivar[f"{mtd} CORRELATION ENERGY"] = mobj.group("corl")
        if mtd not in ["CCSDT-1A", "CCSDT-1B", "CCSDT-2", "CCSDT-3", "CCSDT"]:
            psivar[f"{mtd} DOUBLES ENERGY"] = mobj.group("corl")
        psivar[f"{mtd} TOTAL ENERGY"] = mobj.group("tot")
        module = "ncc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r"Beginning iterative solution of (?P<iterCC>\w+(?:-\d)?) equations" + r"\s*" +
        r'(?P<iterations>.*)' +
        r'^\s*' + r"It\." + r"\s+" + "Correlation Energy" + r".*" +
        r'^\s*(?:-+)\s*' +
        r'^\s*' +
        r'^\s*' + r'(?:\w+(?:-\d)? iterations converged .*?)' +
        r'^\s*' +
        r'^\s*' + r'(?:Total \1 energy:)' + r'\s+' + r'(?P<tot>' + NUMBER + r')\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        mobj2 = re.findall(
            # fmt: off
            r"(\d+)" + r"\s+" + r"(?P<corl>" + DECIMAL + r")\s+" +
                    DECIMAL + r"\s+" + DECIMAL + r"\s+" + DECIMAL + r"\s+" + DECIMAL + r"\s+" + r"(" + DECIMAL + r")?" + r"(" + r"\s+" + DECIMAL + r")?",
            # fmt: on
            mobj.group("iterations"),
        )
        if mobj2:
            logger.debug("matched ncc cc iter mod5", mobj.groupdict(), mobj2[-1])
            mtd = mobj.group("iterCC").upper()
            psivar[f"{mtd} CORRELATION ENERGY"] = mobj2[-1][2]
            if mtd not in ["CCSDT-1A", "CCSDT-1B", "CCSDT-2", "CCSDT-3", "CCSDT"]:
                psivar[f"{mtd} DOUBLES ENERGY"] = mobj2[-1][2]
            psivar[f"{mtd} TOTAL ENERGY"] = mobj.group("tot")
            module = "ncc"

    # Process CC(T)
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:E\(SCF\))' + r'\s+=\s+' + NUMBER + r'\s+a\.u\.\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:E\(CCSD\))' + r'\s+=\s+' + NUMBER + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:E\(CCSD\(T\)\))' + r'\s+=\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) vcc")
        psivar["SCF TOTAL ENERGY"] = mobj.group(1)
        mobj5 = re.search(r"BRUECKNER\s+IBRKNR\s+ON", outtext)
        if not mobj5:
            psivar["CCSD TOTAL ENERGY"] = mobj.group(2)
            psivar["(T) CORRECTION ENERGY"] = Decimal(mobj.group(3)) - Decimal(mobj.group(2))
            psivar["CCSD(T) CORRELATION ENERGY"] = Decimal(mobj.group(3)) - Decimal(mobj.group(1))
            psivar["CCSD(T) TOTAL ENERGY"] = mobj.group(3)
        module = "vcc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:E\(CCSD\))' + r'\s+=\s+' + NUMBER + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:E\(CCSD\(T\)\))' + r'\s+=\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) vcc v2")
        mobj5 = re.search(r"BRUECKNER\s+IBRKNR\s+ON", outtext)
        if mobj5:
            psivar["B(T) CORRECTION ENERGY"] = Decimal(mobj.group(2)) - Decimal(mobj.group(1))
            psivar["BCCD(T) TOTAL ENERGY"] = mobj.group(2)
            psivar["BCCD(T) CORRELATION ENERGY"] = psivar["BCCD CORRELATION ENERGY"] + psivar["B(T) CORRECTION ENERGY"]
        else:
            psivar["CCSD TOTAL ENERGY"] = mobj.group(1)
            psivar["(T) CORRECTION ENERGY"] = Decimal(mobj.group(2)) - Decimal(mobj.group(1))
            psivar["CCSD(T) TOTAL ENERGY"] = mobj.group(2)
        module = "vcc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:E\(SCF\))' + r'\s+=\s*' + NUMBER + r'\s+a\.u\.\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:CCSD energy)' + r'\s+' + NUMBER + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:Total perturbative triples energy:)' + r'\s+' + NUMBER + r'\s*' +
        r'^\s*(?:-+)\s*' +
        r'^\s+' + r'(?:CCSD\(T\) energy)' + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) ecc")
        psivar["SCF TOTAL ENERGY"] = mobj.group(1)
        psivar["CCSD TOTAL ENERGY"] = mobj.group(2)
        psivar["(T) CORRECTION ENERGY"] = mobj.group(3)
        psivar["CCSD(T) CORRELATION ENERGY"] = Decimal(mobj.group(4)) - Decimal(mobj.group(1))
        psivar["CCSD(T) TOTAL ENERGY"] = mobj.group(4)
        module = "ecc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:HF-SCF energy)' + r'\s+' + r"(?P<hf>" + NUMBER + r")" + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:CCSD energy)' + r'\s+' + r"(?P<ccsd>" + NUMBER + r")" + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:E4T  to CCSD\(T\))' + r'\s+' + r"(?P<e4t>" + NUMBER + r")" + r'\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:E4T \+ E5ST)' + r'\s+' + r"(?P<e4te5st>" + NUMBER + r")" + r'\s*' +
        r'(?:.*?)' +
        r'^\s*(?:-+)\s*' +
        r'^\s+' + r'(?:CCSD\(T\) energy)' + r'\s+' + r"(?P<ccsd_t_>" + NUMBER + r")" + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) ecc v2")
        psivar["SCF TOTAL ENERGY"] = mobj.group("hf")
        psivar["CCSD TOTAL ENERGY"] = mobj.group("ccsd")
        psivar["T(CCSD) CORRECTION ENERGY"] = mobj.group("e4t")
        psivar["(T) CORRECTION ENERGY"] = mobj.group("e4te5st")
        psivar["CCSD(T) CORRELATION ENERGY"] = Decimal(mobj.group("ccsd_t_")) - Decimal(mobj.group("hf"))
        psivar["CCSD(T) TOTAL ENERGY"] = mobj.group("ccsd_t_")
        psivar["CCSD+T(CCSD) CORRELATION ENERGY"] = (
            psivar["CCSD CORRELATION ENERGY"] + psivar["T(CCSD) CORRECTION ENERGY"]
        )
        psivar["CCSD+T(CCSD) TOTAL ENERGY"] = psivar["CCSD TOTAL ENERGY"] + psivar["T(CCSD) CORRECTION ENERGY"]
        module = "ecc"

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:CCSD energy)' + r'\s+' + NUMBER + r'\s*' +
        r'^\s*(?:-+)\s*' +
        r'^\s+' + r'(?:CCSD\(T\) energy)' + r'\s+' + NUMBER + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) lamb")
        psivar["CCSD TOTAL ENERGY"] = mobj.group(1)
        psivar["(T) CORRECTION ENERGY"] = Decimal(mobj.group(2)) - Decimal(mobj.group(1))
        psivar["CCSD(T) CORRELATION ENERGY"] = Decimal(mobj.group(2)) - psivar["SCF TOTAL ENERGY"]
        psivar["CCSD(T) TOTAL ENERGY"] = mobj.group(2)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?:CCSD\(T\) contribution:)\s+' + r'(?P<tcorr>' + NUMBER + ')' + r'\s*'
        r'^\s*' + r'(?:CCSD\[T\] contribution:)\s+' + r'(?P<bkttcorr>' + NUMBER + ')' + r'\s*'
        r'^\s*' + r'(?:Total CCSD\(T\) energy:)\s+' + r'(?P<ttot>' + NUMBER + ')' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) ncc")
        psivar["(T) CORRECTION ENERGY"] = mobj.group("tcorr")
        psivar["T(CCSD) CORRECTION ENERGY"] = mobj.group("bkttcorr")
        psivar["CCSD(T) TOTAL ENERGY"] = mobj.group("ttot")
        module = "ncc"

    mobj = re.search(
        # fmt: off
        r'^\s*' + r'(?:CCSD\[T\] correlation energy:)\s+' + r'(?P<bkttcorr>' + NUMBER + ')' + r'\s*'
        r'^\s*' + r'(?:CCSD\(T\) correlation energy:)\s+' + r'(?P<tcorr>' + NUMBER + ')' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd(t) ncc v2")
        psivar["(T) CORRECTION ENERGY"] = mobj.group("tcorr")
        psivar["T(CCSD) CORRECTION ENERGY"] = mobj.group("bkttcorr")
        psivar["CCSD+T(CCSD) TOTAL ENERGY"] = psivar["T(CCSD) CORRECTION ENERGY"] + psivar["CCSD TOTAL ENERGY"]
        psivar["CCSD+T(CCSD) CORRELATION ENERGY"] = (
            psivar["T(CCSD) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]
        )
        psivar["CCSD(T) TOTAL ENERGY"] = psivar["(T) CORRECTION ENERGY"] + psivar["CCSD TOTAL ENERGY"]
        psivar["CCSD(T) CORRELATION ENERGY"] = psivar["(T) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]
        module = "ncc"

    mobj = re.search(
        # fmt: off
        r"^\s*" + r"(?:Lambda-CCSD iterations converged .*)" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:CCSD\[T\] correlation energy:)\s+" + r"(?P<bkttcorr>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\(T\) correlation energy:)\s+" + r"(?P<tcorr>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\[T\]_L correlation energy:)\s+" + r"(?P<abkttcorr>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\(T\)_L correlation energy:)\s+" + r"(?P<atcorr>" + NUMBER + r")" + r"\s*" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:Non-iterative calculation of CCSD\(T\)_L .*)" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:Total CCSD\(T\)_L energy:)\s+" + r"(?P<accsdttot>" + NUMBER + r")" + r"\s*$",
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched a-ccsd(t) ncc")
        psivar["(T) CORRECTION ENERGY"] = mobj.group("tcorr")
        psivar["T(CCSD) CORRECTION ENERGY"] = mobj.group("bkttcorr")
        psivar["A-(T) CORRECTION ENERGY"] = mobj.group("atcorr")
        psivar["CCSD+T(CCSD) TOTAL ENERGY"] = psivar["T(CCSD) CORRECTION ENERGY"] + psivar["CCSD TOTAL ENERGY"]
        psivar["CCSD+T(CCSD) CORRELATION ENERGY"] = (
            psivar["T(CCSD) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]
        )
        psivar["CCSD(T) TOTAL ENERGY"] = psivar["(T) CORRECTION ENERGY"] + psivar["CCSD TOTAL ENERGY"]
        psivar["CCSD(T) CORRELATION ENERGY"] = psivar["(T) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]
        psivar["A-CCSD(T) TOTAL ENERGY"] = mobj.group("accsdttot")
        psivar["A-CCSD(T) CORRELATION ENERGY"] = psivar["A-(T) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]
        module = "ncc"

    mobj = re.search(
        # fmt: off
        r"^\s*" + r"(?:A miracle come to pass. The CC iterations have converged.)" + r"\s*" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:HF-SCF energy          )\s+" + r"(?P<hf>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:MP2 correlation energy )\s+" + r"(?P<mp2corl>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:MP2 energy             )\s+" + r"(?P<mp2>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD correlation energy)\s+" + r"(?P<ccsdcorl>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD energy            )\s+" + r"(?P<ccsd>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:Delta ET to CCSD\[T\]_L)\s+" + r"(?P<abkttcorr>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\[T\]_L energy     )\s+" + r"(?P<accsdbktt>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:Delta ET to CCSD\(T\)_L)\s+" + r"(?P<atcorr>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\(T\)_L energy     )\s+" + r"(?P<accsdt>" + NUMBER + r")" + r"\s*" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:Non-iterative perturbative treatment of triple)" + r"\s*" +
        r"^\s*" + r"(?:excitations using the CCSD\(T\)_L method:)" + r"\s*" +
        r'(?:.*?)' +
        r"^\s*" + r"(?:HF-SCF energy)\s+" + r"(?P<hf_again>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:MP2 correlation energy)\s+" + r"(?P<mp2corl_again>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:MP2 energy)\s+" + r"(?P<mp2_again>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD correlation energy)\s+" + r"(?P<lccsdcorl>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD energy)\s+" + r"(?P<lccsd>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:Delta ET to CCSD\(T\)_L)\s+" + r"(?P<atcorr_again>" + NUMBER + r")" + r"\s*" +
        r"^\s*" + r"(?:CCSD\(T\)_L energy)\s+" + r"(?P<laccsdt>" + NUMBER + r")" + r"\s*$",
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched a-ccsd(t) ecc")
        psivar["HF TOTAL ENERGY"] = mobj.group("hf")
        psivar["MP2 CORRELATION ENERGY"] = mobj.group("mp2corl")
        psivar["CCSD CORRELATION ENERGY"] = mobj.group("ccsdcorl")
        psivar["A-(T) CORRECTION ENERGY"] = mobj.group("atcorr")
        psivar["A-CCSD(T) TOTAL ENERGY"] = mobj.group("accsdt")
        psivar["A-CCSD(T) CORRELATION ENERGY"] = psivar["A-(T) CORRECTION ENERGY"] + psivar["CCSD CORRELATION ENERGY"]

        mobj3 = re.search(r"SCF reference function:  (R|U)HF", outtext)
        if mobj3:
            psivar["CCSD SINGLES ENERGY"] = Decimal("0.0")
        module = "ecc"

    mobj = re.search(
        # fmt: off
           r"^\s*" + r"(?:E\(CCSD\))" + r"\s+=\s+" + r"(?P<ccsdtot>" + NUMBER + ")" + r"\s*" +
           r"^\s*" + r"(?:E\(CCSD \+ T\(CCSD\)\))" + r"\s*=\s*" + r"(?P<ccsdtccsdtot>" + NUMBER + ")" + r"\s*$",
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsd+t(ccsd) vcc")
        psivar["CCSD TOTAL ENERGY"] = mobj.group("ccsdtot")
        psivar["CCSD+T(CCSD) TOTAL ENERGY"] = mobj.group("ccsdtccsdtot")
        psivar["CCSD+T(CCSD) CORRELATION ENERGY"] = psivar["CCSD+T(CCSD) TOTAL ENERGY"] - psivar["SCF TOTAL ENERGY"]
        mobj3 = re.search(r"Reference function is (R|U)HF Hartree-Fock", outtext)
        if mobj3:
            psivar["CCSD SINGLES ENERGY"] = Decimal("0.0")
        module = "vcc"

    mobj = re.search(
        # fmt: off
        r'^\s*' + r'(?:CCSDT\[Q\] correlation energy:)\s+' + r'(?P<bkttcorr>' + NUMBER + ')' + r'\s*'
        r'^\s*' + r'(?:CCSDT\(Q\) correlation energy:)\s+' + r'(?P<tcorr>' + NUMBER + ')' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched ccsdt(q) ncc")
        psivar["(Q) CORRECTION ENERGY"] = mobj.group("tcorr")
        psivar["[Q] CORRECTION ENERGY"] = mobj.group("bkttcorr")
        psivar["CCSDT(Q) TOTAL ENERGY"] = psivar["(Q) CORRECTION ENERGY"] + psivar["CCSDT TOTAL ENERGY"]
        psivar["CCSDT(Q) CORRELATION ENERGY"] = psivar["(Q) CORRECTION ENERGY"] + psivar["CCSDT CORRELATION ENERGY"]
        module = "ncc"

    # Process DBOC
    mobj = re.search(
        # fmt: off
        r'^\s*' + r'(?:The total diagonal Born-Oppenheimer correction \(DBOC\) is:)\s+' +
        r'(?P<dboc>' + NUMBER + ')' + r'\s*a.u.\s*',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:
        logger.debug("matched dboc ecc")
        psivar["CCSD DBOC ENERGY"] = mobj.group("dboc")
        module = "ecc"

    # Process SCS-CC
    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?P<fullCC>(?P<iterCC>CC(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s*' + r'(?:@CCENRG-I, Correlation energies.)' + r'\s+(?:ECCAA)\s+' + NUMBER + r'\s*' +
        r'^\s+(?:ECCBB)\s+' + NUMBER + r'\s*' +
        r'^\s+(?:ECCAB)\s+' + NUMBER + r'\s*' +
        r'^\s+(?:Total)\s+' + NUMBER + r'\s*',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:  # PRINT=2 to get SCS-CC components
        logger.debug("matched scscc")
        if float(mobj.group(4)) == 0.0:
            ss = 2 * Decimal(mobj.group(3))
        else:
            ss = Decimal(mobj.group(3)) + Decimal(mobj.group(4))

        mobj5 = re.search(r"BRUECKNER\s+IBRKNR\s+ON", outtext)
        if not mobj5:
            if not (
                re.search(r"executable xvcc finished", outtext)
                and re.search(r"The reference state is a ROHF wave function.", outtext)
            ):
                psivar["%s SAME-SPIN CORRELATION ENERGY" % (mobj.group("iterCC"))] = ss
            psivar["%s OPPOSITE-SPIN CORRELATION ENERGY" % (mobj.group("iterCC"))] = mobj.group(5)
            psivar["%s CORRELATION ENERGY" % (mobj.group("iterCC"))] = mobj.group(6)

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?P<fullCC>(?P<iterCC>CC(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'Amplitude equations converged in' + r'\s*\d+\s*' + r'iterations.\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'The AA contribution to the correlation energy is:\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'The BB contribution to the correlation energy is:\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'The AB contribution to the correlation energy is:\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'The total correlation energy is\s+' + NUMBER + r'\s+a.u.\s*' +
        r'(?:.*?)' +
        # r'^\s+' + r'The CC iterations have converged.' + r'\s*$',
        r'^\s+' + r'(?:A miracle come to pass. )?' + r'The CC iterations have converged.' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:  # PRINT=2 to get SCS components
        logger.debug("matched scscc2")
        iterCC = mobj.group("iterCC")
        mobj3 = re.search(r"The reference state is a ROHF wave function.", outtext)
        mobj4 = re.search(r"executable xvcc finished", outtext)
        if mobj4:  # vcc
            if not (iterCC == "CCD" and mobj3):
                # uncertain if ROHF CCD correct
                psivar[f"{iterCC} OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(5)
            if not mobj3:
                psivar[f"{iterCC} SAME-SPIN CORRELATION ENERGY"] = Decimal(mobj.group(3)) + Decimal(mobj.group(4))
        else:  # ecc
            psivar[f"{iterCC} SAME-SPIN CORRELATION ENERGY"] = Decimal(mobj.group(3)) + Decimal(mobj.group(4))
            if not mobj3:
                psivar[f"{iterCC} OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group(5)
        psivar[f"{iterCC} CORRELATION ENERGY"] = mobj.group(6)

    mobj = re.search(
        # fmt: off
        #r'^\s+' + r'(?P<fullCC>(?P<iterCC>L?CC(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        # better one for LCC and one for CC, right?
        r'^\s+' + r'(?P<fullCC>(?P<iterCC>LCC(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'Amplitude equations converged in' + r'\s*\d+\s*' + r'iterations.\s*' +
        r'(?:.*?)' +
         r'^\s+' + r'The AA contribution to the correlation energy is:\s+' + r"(?P<AA>" + NUMBER + r")" + r'\s+a.u.\s*' +
        r'(^\s+' + r'The BB contribution to the correlation energy is:\s+' + r"(?P<BB>" + NUMBER + r")" + r'\s+a.u.\s*' + r")?" +
         r'^\s+' + r'The AB contribution to the correlation energy is:\s+' + r"(?P<AB>" + NUMBER + r")" + r'\s+a.u.\s*' +
        r'^\s+' + r'The total correlation energy is\s+' + r"(?P<corl>" + NUMBER + r")" + r'\s+a.u.\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'(?:A miracle come to pass. )?' + r'The CC iterations have converged.' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:  # PRINT=2 to get SCS components
        logger.debug("matched scslccd")
        mobj3 = re.search(r"The reference state is a ROHF wave function.", outtext)
        mobj4 = re.search(r"executable xvcc finished", outtext)
        iterCC = mobj.group("iterCC")
        if mobj4:  # vcc
            if mobj.group("BB"):
                aabb = Decimal(mobj.group("AA")) + Decimal(mobj.group("BB"))
            else:
                aabb = Decimal("2") * Decimal(mobj.group("AA"))
            psivar[f"{iterCC} OPPOSITE-SPIN CORRELATION ENERGY"] = mobj.group("AB")
            if not mobj3:
                psivar[f"{iterCC} SAME-SPIN CORRELATION ENERGY"] = aabb
        psivar["%s CORRELATION ENERGY" % (mobj.group("iterCC"))] = mobj.group("corl")

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'(?P<fullCC>(?P<iterCC>CC(?:\w+))(?:\(T\))?)' + r'\s+(?:energy will be calculated.)\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'Amplitude equations converged in' + r'\s*\d+\s*' + r'iterations.\s*' +
        r'(?:.*?)' +
        r'^\s+' + r'The AA contribution to the correlation energy is:\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'The AB contribution to the correlation energy is:\s+' + NUMBER + r'\s+a.u.\s*' +
        r'^\s+' + r'The total correlation energy is\s+' + NUMBER + r'\s+a.u.\s*' +
        r'(?:.*?)' +
        # r'^\s+' + r'The CC iterations have converged.' + r'\s*$',
        r'^\s+' + r'(?:A miracle come to pass. )?' + r'The CC iterations have converged.' + r'\s*$',
        # fmt: on
        outtext,
        re.MULTILINE | re.DOTALL,
    )
    if mobj:  # PRINT=2 to get SCS components
        logger.debug("matched scscc rhf", mobj.groups())
        psivar["%s SAME-SPIN CORRELATION ENERGY" % (mobj.group("iterCC"))] = 2 * Decimal(mobj.group(3))
        psivar["%s OPPOSITE-SPIN CORRELATION ENERGY" % (mobj.group("iterCC"))] = mobj.group(4)
        psivar["%s CORRELATION ENERGY" % (mobj.group("iterCC"))] = mobj.group(5)

    # Process gradient
    mobj = re.search(
        # fmt: off
        r'\s+' + r'Molecular gradient' + r'\s*' +
        r'\s+' + r'------------------' + r'\s*' +
        r'\s+' + r'\n' +
        r'(?:(?:\s+[A-Z]+\s*#\d+\s+[xyz]\s+[-+]?\d+\.\d+\s*\n)+)' +  # optional, it seems
        r'\n\n' +  # optional, it seems
        r'((?:\s+[A-Z]+\s*#\d+\s+\d?\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' +
        r'\n\n' +
        r'\s+' + 'Molecular gradient norm',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched molgrad")
        atoms = []
        psivar_grad = []
        for line in mobj.group(1).splitlines():
            lline = line.split()
            atoms.append(lline[0])
            # psivar_gradient.append([Decimal(lline[-3]), Decimal(lline[-2]), Decimal(lline[-1])])
            psivar_grad.append([float(lline[-3]), float(lline[-2]), float(lline[-1])])

    # Process geometry
    mobj = re.search(
        # fmt: off
        # r'\s+(?:-+)\s*' +
        # r'^\s+' + r'Z-matrix   Atomic            Coordinates (in bohr)' + r'\s*' +
        r'^\s+' + r'Symbol    Number           X              Y              Z' + r'\s*' +
        r'^\s+(?:-+)\s*' +
        r'((?:\s+[A-Z]+\s+([0-9]+|\*\*\*)\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s+[-+]?\d+\.\d+\s*\n)+)' +  # allows ghosts
        r'^\s+(?:-+)\s*',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched geom")
        molxyz = "%d bohr\n\n" % len(mobj.group(1).splitlines())
        for line in mobj.group(1).splitlines():
            lline = line.split()
            if lline[1] == "***":
                tag = "@Xe"  # potentially dangerous bypass
            else:
                tag = lline[0]
            molxyz += "%s %16s %16s %16s\n" % (tag, lline[-3], lline[-2], lline[-1])
        # Rather a dinky Molecule as no ghost, charge, or multiplicity
        psivar_coord = Molecule(
            validate=False,
            **qcel.molparse.to_schema(
                qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"], dtype=2
            ),
        )

    # Process atom geometry
    mobj = re.search(r"^\s+" + r"@GETXYZ-I,     1 atoms read from ZMAT." + r"\s*$", outtext, re.MULTILINE)
    mobj2 = re.search(
        r"^([A-Z]+)#1" + r"\s+" + NUMBER + r"\s+" + NUMBER + r"\s+" + NUMBER + r"\s*$", outtext, re.MULTILINE
    )
    if mobj and mobj2:
        logger.debug("matched atom2")  # unsavory for when atom never printed except for basis file
        # Dinky Molecule
        molxyz = "1 bohr\n\n%s 0.0 0.0 0.0\n" % (mobj2.group(1))
        psivar_coord = Molecule(
            validate=False,
            **qcel.molparse.to_schema(
                qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"], dtype=2
            ),
        )

    mobj = re.search(
        # fmt: off
        r'^\s+' + r'@GETXYZ-I,     1 atoms read from ZMAT.' + r'\s*' +
        r'^\s+' + r'[0-9]+\s+([A-Z]+)\s+[0-9]+\s+' + NUMBER + r'\s*',
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched atom")
        # Dinky Molecule
        molxyz = "1 bohr\n\n%s 0.0 0.0 0.0\n" % (mobj.group(1))
        psivar_coord = Molecule(
            validate=False,
            **qcel.molparse.to_schema(
                qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"], dtype=2
            ),
        )

    # Process error codes
    mobj = re.search(
        # fmt: off
        r"^\s*" + r"--executable " + r"(?P<c4exe>\w+)" + r" finished with status" + r"\s+" + r"(?P<errcode>[1-9][0-9]*)",
        # fmt: on
        outtext,
        re.MULTILINE,
    )
    if mobj:
        logger.debug("matched error")
        # psivar['CFOUR ERROR CODE'] = mobj.group(2)
        c4exe = mobj.group("c4exe")
        errcode = int(mobj.group("errcode"))
        if errcode != 0:
            error += f"--executable {c4exe} finished with status {errcode}"
            if c4exe in ["xvcc", "xecc", "xncc"]:
                module = c4exe[1:]

    # Process CURRENT energies (TODO: needs better way)
    if "SCF TOTAL ENERGY" in psivar:
        psivar["CURRENT REFERENCE ENERGY"] = psivar["SCF TOTAL ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["SCF TOTAL ENERGY"]
        psivar["HF TOTAL ENERGY"] = psivar["SCF TOTAL ENERGY"]

    if "MP2 TOTAL ENERGY" in psivar and "MP2 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP2 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP2 TOTAL ENERGY"]

    if "MP3 TOTAL ENERGY" in psivar and "MP3 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP3 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP3 TOTAL ENERGY"]

    if "MP4(SDQ) TOTAL ENERGY" in psivar and "MP4(SDQ) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP4(SDQ) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP4(SDQ) TOTAL ENERGY"]

    if "MP4 TOTAL ENERGY" in psivar and "MP4 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["MP4 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["MP4 TOTAL ENERGY"]

    if "CISD TOTAL ENERGY" in psivar and "CISD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CISD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CISD TOTAL ENERGY"]

    if "QCISD TOTAL ENERGY" in psivar and "QCISD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["QCISD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["QCISD TOTAL ENERGY"]

    if "QCISD(T) TOTAL ENERGY" in psivar and "QCISD(T) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["QCISD(T) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["QCISD(T) TOTAL ENERGY"]

    if "LCCD TOTAL ENERGY" in psivar and "LCCD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["LCCD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["LCCD TOTAL ENERGY"]

    if "LCCSD TOTAL ENERGY" in psivar and "LCCSD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["LCCSD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["LCCSD TOTAL ENERGY"]

    #    if ('%s TOTAL ENERGY' % (mobj.group('fullCC')) in psivar) and \
    #       ('%s CORRELATION ENERGY' % (mobj.group('fullCC')) in psivar):
    #        psivar['CURRENT CORRELATION ENERGY'] = psivar['%s CORRELATION ENERGY' % (mobj.group('fullCC')]
    #        psivar['CURRENT ENERGY'] = psivar['%s TOTAL ENERGY' % (mobj.group('fullCC')]

    if "CC2 TOTAL ENERGY" in psivar and "CC2 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CC2 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CC2 TOTAL ENERGY"]

    if "CCD TOTAL ENERGY" in psivar and "CCD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCD TOTAL ENERGY"]

    if "BCCD TOTAL ENERGY" in psivar and "BCCD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["BCCD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["BCCD TOTAL ENERGY"]

    if "CCSD TOTAL ENERGY" in psivar and "CCSD CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD TOTAL ENERGY"]

    if "CCSD+T(CCSD) TOTAL ENERGY" in psivar and "CCSD+T(CCSD) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD+T(CCSD) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD+T(CCSD) TOTAL ENERGY"]

    if "CCSD(T) TOTAL ENERGY" in psivar and "CCSD(T) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSD(T) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSD(T) TOTAL ENERGY"]

    if "A-CCSD(T) TOTAL ENERGY" in psivar and "A-CCSD(T) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["A-CCSD(T) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["A-CCSD(T) TOTAL ENERGY"]

    if "BCCD(T) TOTAL ENERGY" in psivar and "BCCD(T) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["BCCD(T) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["BCCD(T) TOTAL ENERGY"]

    if "CC3 TOTAL ENERGY" in psivar and "CC3 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CC3 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CC3 TOTAL ENERGY"]

    if "CCSDT TOTAL ENERGY" in psivar and "CCSDT CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT TOTAL ENERGY"]

    if "CCSDT-1A TOTAL ENERGY" in psivar and "CCSDT-1A CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT-1A CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT-1A TOTAL ENERGY"]

    if "CCSDT-1B TOTAL ENERGY" in psivar and "CCSDT-1B CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT-1B CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT-1B TOTAL ENERGY"]

    if "CCSDT-2 TOTAL ENERGY" in psivar and "CCSDT-2 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT-2 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT-2 TOTAL ENERGY"]

    if "CCSDT-3 TOTAL ENERGY" in psivar and "CCSDT-3 CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT-3 CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT-3 TOTAL ENERGY"]

    if "CCSDT(Q) TOTAL ENERGY" in psivar and "CCSDT(Q) CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDT(Q) CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDT(Q) TOTAL ENERGY"]

    if "CCSDTQ TOTAL ENERGY" in psivar and "CCSDTQ CORRELATION ENERGY" in psivar:
        psivar["CURRENT CORRELATION ENERGY"] = psivar["CCSDTQ CORRELATION ENERGY"]
        psivar["CURRENT ENERGY"] = psivar["CCSDTQ TOTAL ENERGY"]

    psivar[f"N ATOMS"] = len(psivar_coord.symbols)

    return psivar, psivar_coord, psivar_grad, version, module, error


def harvest(in_mol: Molecule, method: str, c4out, **largs):
    """Parses all the pieces of output from Cfour: the stdout in
    *c4out* and the contents of various scratch files like GRD stored
    in their namesake keys in *largs*. Since all Cfour output uses
    its own orientation and atom ordering for the given molecule,
    a qcdb.Molecule *in_mol*, if supplied, is used to transform the
    Cfour output back into consistency with *in_mol*.

    """
    # Collect results from output file and subsidiary files
    qcvars, out_mol, outGrad, version, module, error = harvest_output(c4out)

    if largs.get("GRD"):
        grdMol, grdGrad = harvest_GRD(largs["GRD"])
    else:
        grdMol, grdGrad = None, None

    if largs.get("FCMFINAL"):
        fcmHess = load_hessian(largs["FCMFINAL"], dtype="fcmfinal")
        if np.count_nonzero(fcmHess) == 0:
            fcmHess = None
    else:
        fcmHess = None

    if largs.get("DIPOL"):
        dipolDip = harvest_DIPOL(largs["DIPOL"])
    else:
        dipolDip = None

    # Sometimes the hierarchical setting of CURRENT breaks down
    if method == "ccsd+t(ccsd)":
        qcvars["CURRENT CORRELATION ENERGY"] = qcvars["CCSD+T(CCSD) CORRELATION ENERGY"]
        qcvars["CURRENT ENERGY"] = qcvars["CCSD+T(CCSD) TOTAL ENERGY"]

    if fcmHess is not None and method in ["hf", "scf"]:
        # MP2 available in HF Hessian so need to counteract
        qcvars.pop("CURRENT CORRELATION ENERGY")
        qcvars["CURRENT ENERGY"] = qcvars["HF TOTAL ENERGY"]

    # Reconcile the coordinate information: several cases
    #   Case                            p4Mol   GRD      Check consistency           Apply orientation?     ReturnMol (1-19-2014)
    #   sp with mol thru cfour {}       None    None              outMol             N.C.                   outMol
    #   opt with mol thru cfour {}      None    grdMol            outMol && grdMol   N.C.                   grdMol
    #   sp with mol thru molecule {}    p4Mol   None     p4Mol && outMol             p4Mol <-- outMol       p4Mol (same as input arg)
    #   opt with mol thru molecule {}   p4Mol   grdMol   p4Mol && outMol && grdMol   p4Mol <-- grdMol       p4Mol (same as input arg)
    # Jul 2021: above describes longtime orientation strategy. Now, mol through cfour {} no longer allowed, and fix_* signal whether input (T) or cfour native (F) frames for returned data.

    if out_mol:
        if grdMol:
            if abs(out_mol.nuclear_repulsion_energy() - grdMol.nuclear_repulsion_energy()) > 1.0e-3:
                raise ValueError(
                    f"""CFOUR outfile (NRE: {out_mol.nuclear_repulsion_energy()} inconsistent with CFOUR GRD (NRE: {grdMol.nuclear_repulsion_energy()})."""
                )
        if in_mol:
            if abs(out_mol.nuclear_repulsion_energy() - in_mol.nuclear_repulsion_energy()) > 1.0e-3:
                raise ValueError(
                    f"""CFOUR outfile (NRE: {out_mol.nuclear_repulsion_energy()}) inconsistent with AtomicInput.molecule (NRE: {in_mol.nuclear_repulsion_energy()})."""
                )
    else:
        raise ValueError("""No coordinate information extracted from CFOUR output.""")

    # Set up array reorientation object(s)
    if in_mol and out_mol and grdMol:
        # Jul 2021: apparently GRD and FCMFINAL can have different atom orderings :-)

        _, data = grdMol.align(out_mol, atoms_map=False, mols_align=True, verbose=0)
        g2o_mill = data["mill"]

        oriCoord = g2o_mill.align_coordinates(grdMol.geometry)
        oriGrad = g2o_mill.align_gradient(np.array(grdGrad))

        if dipolDip is None:
            oriDip = None
        else:
            oriDip = g2o_mill.align_vector(dipolDip)

        if fcmHess is None:
            oriHess = None
        else:
            oriHess = fcmHess

        # Frame considerations
        if in_mol.fix_com and in_mol.fix_orientation:
            # Impose input frame if important as signalled by fix_*=T
            return_mol = in_mol
            _, data = out_mol.align(in_mol, atoms_map=False, mols_align=True, generic_ghosts=True, verbose=0)
            mill = data["mill"]

        else:
            return_mol, _ = in_mol.align(out_mol, atoms_map=False, mols_align=True, generic_ghosts=True, verbose=0)
            mill = qcel.molutil.compute_scramble(
                len(in_mol.symbols), do_resort=False, do_shift=False, do_rotate=False, do_mirror=False
            )  # identity AlignmentMill

        #        _, data = out_mol.align(in_mol, atoms_map=False, mols_align=True, verbose=0)
        #        o2i_mill = data["mill"]
        #
        #        oriCoord = o2i_mill.align_coordinates(oriCoord)
        #        oriGrad = o2i_mill.align_gradient(oriGrad)
        #        if oriDip is not None:
        #            oriDip = o2i_mill.align_vector(oriDip)
        #        if oriHess is not None:
        #            oriHess = o2i_mill.align_hessian(oriHess)

        oriCoord = mill.align_coordinates(oriCoord)
        oriGrad = mill.align_gradient(oriGrad)
        if oriDip is not None:
            oriDip = mill.align_vector(oriDip)
        if oriHess is not None:
            oriHess = mill.align_hessian(oriHess)

    elif in_mol and out_mol:
        # TODO watch out - haven't seen atom_map=False yet

        if in_mol.fix_com and in_mol.fix_orientation:
            # Impose input frame if important as signalled by fix_*=T
            return_mol = in_mol
            _, data = out_mol.align(in_mol, atoms_map=True, mols_align=True, generic_ghosts=True, verbose=0)
            mill = data["mill"]

        else:
            return_mol, _ = in_mol.align(out_mol, atoms_map=False, mols_align=True, generic_ghosts=True, verbose=0)
            mill = qcel.molutil.compute_scramble(
                len(in_mol.symbols), do_resort=False, do_shift=False, do_rotate=False, do_mirror=False
            )  # identity AlignmentMill

        oriCoord = mill.align_coordinates(out_mol.geometry)  # (np_out=True))
        oriGrad = None
        oriHess = None  # I don't think we ever get FCMFINAL w/o GRAD
        if dipolDip is None:
            oriDip = None
        else:
            oriDip = mill.align_vector(np.array(dipolDip))
        # p4c4 = OrientMols(in_mol, out_mol)
        # oriCoord = p4c4.transform_coordinates2(out_mol)
        # oriGrad = None
        # oriDip = None if dipolDip is None else p4c4.transform_vector(dipolDip)

    elif out_mol:
        oriGrad = None
        oriHess = None
        oriDip = None if dipolDip is None else dipolDip

    # not sure of purpose but it interferes now that return_mol overwrites atres.mol
    # return_mol = None if in_mol else grdMol

    if oriDip is not None:
        qcvars["CURRENT DIPOLE"] = oriDip
        oriDip *= qcel.constants.dipmom_au2debye
        # outPsivar["CURRENT DIPOLE X"] = oriDip[0]
        # outPsivar["CURRENT DIPOLE Y"] = oriDip[1]
        # outPsivar["CURRENT DIPOLE Z"] = oriDip[2]
        # outPsivar['CURRENT DIPOLE X'] = str(oriDip[0] * psi_dipmom_au2debye)
        # outPsivar['CURRENT DIPOLE Y'] = str(oriDip[1] * psi_dipmom_au2debye)
        # outPsivar['CURRENT DIPOLE Z'] = str(oriDip[2] * psi_dipmom_au2debye)

    if oriGrad is not None:
        return_grad = oriGrad
    elif grdGrad is not None:
        return_grad = grdGrad
    else:
        return_grad = None

    if oriHess is not None:
        return_hess = oriHess
    else:
        return_hess = None

    # if oriCoord is not None:
    #     retCoord = oriCoord
    # else:
    #     retCoord = None

    return qcvars, return_hess, return_grad, return_mol, version, module, error


def harvest_GRD(grd):
    """Parses the contents *grd* of the Cfour GRD file into the gradient
    array and coordinate information. The coordinate info is converted
    into a rather dinky Molecule (no charge, multiplicity, or fragment),
    but this is these coordinates that govern the reading of molecule
    orientation by Cfour. Return qcel.models.Molecule and gradient array.

    """
    grd = grd.splitlines()
    Nat = int(grd[0].split()[0])
    molxyz = f"{Nat} bohr\n\n"

    grad = []
    for at in range(Nat):
        mline = grd[at + 1].split()

        # "@Xe" is potentially dangerous bypass for ghosts
        el = "@Xe" if int(float(mline[0])) == 0 else qcel.periodictable.to_E(int(float(mline[0])))
        molxyz += "%s %16s %16s %16s\n" % (el, mline[-3], mline[-2], mline[-1])
        lline = grd[at + 1 + Nat].split()
        grad.append([float(lline[-3]), float(lline[-2]), float(lline[-1])])
    grad = np.array(grad).reshape((-1, 3))

    mol = Molecule(
        validate=False,
        **qcel.molparse.to_schema(
            qcel.molparse.from_string(molxyz, dtype="xyz+", fix_com=True, fix_orientation=True)["qm"], dtype=2
        ),
    )

    return mol, grad


def harvest_DIPOL(dipol):
    """Parses the contents *dipol* of the Cfour DIPOL file into a dipol vector."""
    dipol = dipol.splitlines()
    lline = dipol[0].split()
    dip = np.array([float(lline[0]), float(lline[1]), float(lline[2])])

    # return None if empty else dip
    return dip



import os
import qcengine as qcng
import qcelemental as qcel

os.environ["TERACHEM_PBS_HOST"] = "127.0.0.1"
os.environ["TERACHEM_PBS_PORT"] = "11111"

prog = qcng.get_program("terachem_pbs")


mol = qcel.models.Molecule.from_data(
    """
    O  0.0  0.000  -0.129
    H  0.0 -1.494  1.027
    H  0.0  1.494  1.027
"""
)

inp = qcel.models.AtomicInput(
    molecule=mol,
    driver="energy",
    model={"method": "pbe0", "basis": "6-31g"},
)
ret = prog.compute(inp)
print(ret)